
from .main import Slugify, UniqueSlugify
from .alt_translates import *


slugify = Slugify()
unique_slugify = UniqueSlugify()
slugify_unicode = Slugify(translate=None)

slugify_url = Slugify()
slugify_url.to_lower = True
slugify_url.stop_words = ('a', 'an', 'the')
slugify_url.max_length = 200

slugify_filename = Slugify()
slugify_filename.separator = '_'
slugify_filename.safe_chars = '-.'
slugify_filename.max_length = 255

slugify_ru = Slugify(pretranslate=CYRILLIC)
slugify_de = Slugify(pretranslate=GERMAN)
slugify_el = Slugify(pretranslate=GREEK)


# Legacy code
def deprecate_init(Klass):
    class NewKlass(Klass):
        def __init__(self, *args, **kwargs):
            import warnings
            warnings.simplefilter('once')
            warnings.warn("'slugify.get_slugify' is deprecated; use 'slugify.Slugify' instead.",
                          DeprecationWarning, stacklevel=2)
            super(NewKlass, self).__init__(*args, **kwargs)
    return NewKlass

# get_slugify was deprecated in 2014, march 31
get_slugify = deprecate_init(Slugify)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class TerminalFilterEntry(BaseModel):
    name: str
    """The name of the filter."""

    regex: str
    """The regular expression to match. Use [JavaScript regular expressions](https://developer.mozilla.org/en/docs/Web/JavaScript/Guide/Regular_Expressions)."""


DEFAULT_TERMINAL_FILTERS = [
    TerminalFilterEntry(
        name="Suppress temperature messages",
        regex=r"(Send: (N\d+\s+)?M105)|(Recv:\s+(ok\s+([PBN]\d+\s+)*)?([BCLPR]|T\d*):-?\d+)",
    ),
    TerminalFilterEntry(
        name="Suppress SD status messages",
        regex=r"(Send: (N\d+\s+)?M27)|(Recv: SD printing byte)|(Recv: Not SD printing)",
    ),
    TerminalFilterEntry(
        name="Suppress position messages",
        regex=r"(Send:\s+(N\d+\s+)?M114)|(Recv:\s+(ok\s+)?X:[+-]?([0-9]*[.])?[0-9]+\s+Y:[+-]?([0-9]*[.])?[0-9]+\s+Z:[+-]?([0-9]*[.])?[0-9]+\s+E\d*:[+-]?([0-9]*[.])?[0-9]+).*",
    ),
    TerminalFilterEntry(name="Suppress wait responses", regex=r"Recv: wait"),
    TerminalFilterEntry(
        name="Suppress processing responses",
        regex=r"Recv: (echo:\s*)?busy:\s*processing",
    ),
]



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import sys

from octoprint.util.commandline import CommandlineCaller, CommandlineError

from ..exceptions import CannotUpdateOffline, ConfigurationInvalid, UpdateError


def _get_caller(log_cb=None):
    def _log_call(*lines):
        _log(lines, prefix=" ", stream="call")

    def _log_stdout(*lines):
        _log(lines, prefix=">", stream="stdout")

    def _log_stderr(*lines):
        _log(lines, prefix="!", stream="stderr")

    def _log(lines, prefix=None, stream=None):
        if log_cb is None:
            return
        log_cb(lines, prefix=prefix, stream=stream)

    caller = CommandlineCaller()
    if log_cb is not None:
        caller.on_log_call = _log_call
        caller.on_log_stdout = _log_stdout
        caller.on_log_stderr = _log_stderr
    return caller


def can_perform_update(target, check, online=True):
    import os

    script_configured = bool("update_script" in check and check["update_script"])

    folder = None
    if "update_folder" in check:
        folder = check["update_folder"]
    elif "checkout_folder" in check:
        folder = check["checkout_folder"]
    folder_configured = bool(folder and os.path.isdir(folder))

    return (
        script_configured
        and folder_configured
        and (online or check.get("offline", False))
    )


def perform_update(target, check, target_version, log_cb=None, online=True, force=False):
    logger = logging.getLogger("octoprint.plugins.softwareupdate.updaters.update_script")

    if not online and not check("offline", False):
        raise CannotUpdateOffline()

    if not can_perform_update(target, check):
        raise ConfigurationInvalid(
            "checkout_folder and update_folder are missing for update target %s, one is needed"
            % target
        )

    update_script = check["update_script"]
    update_branch = check.get("update_branch", "")
    force_exact_version = check.get("force_exact_version", False)
    folder = check.get(
        "update_folder", check.get("checkout_folder")
    )  # either should be set, tested above
    pre_update_script = check.get("pre_update_script", None)
    post_update_script = check.get("post_update_script", None)

    caller = _get_caller(log_cb=log_cb)

    ### pre update

    if pre_update_script is not None:
        logger.debug(f"Target: {target}, running pre-update script: {pre_update_script}")
        try:
            caller.checked_call(pre_update_script, cwd=folder)
        except CommandlineError as e:
            logger.exception(
                "Target: %s, error while executing pre update script, got returncode %r"
                % (target, e.returncode)
            )

    ### update

    try:
        update_command = update_script.format(
            python=sys.executable,
            folder=folder,
            target=target_version,
            branch=update_branch,
            force="true" if force_exact_version else "false",
        )

        logger.debug(f"Target {target}, running update script: {update_command}")

        caller.checked_call(update_command, cwd=folder)
    except CommandlineError as e:
        logger.exception(
            "Target: %s, error while executing update script, got returncode %r"
            % (target, e.returncode)
        )
        raise UpdateError(
            "Error while executing update script for %s", (e.stdout, e.stderr)
        )

    ### post update

    if post_update_script is not None:
        logger.debug(
            "Target: {}, running post-update script {}...".format(
                target, post_update_script
            )
        )
        try:
            caller.checked_call(post_update_script, cwd=folder)
        except CommandlineError as e:
            logger.exception(
                "Target: %s, error while executing post update script, got returncode %r"
                % (target, e.returncode)
            )

    return "ok"



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import errno
import logging

import requests.exceptions
import serial
import tornado.websocket
from flask import jsonify
from flask_babel import gettext

import octoprint.plugin
from octoprint.util import get_fully_qualified_classname as fqcn  # noqa: F401
from octoprint.util.version import (
    get_octoprint_version_string,
    is_released_octoprint_version,
)

SENTRY_URL_SERVER = (
    "https://1af1a737d9684867859b6baec346addd@o118517.ingest.sentry.io/1373987"
)
SENTRY_URL_COREUI = (
    "https://9823e9176ee54cf385e31cbbef50dda7@o118517.ingest.sentry.io/1374096"
)

SETTINGS_DEFAULTS = {
    "enabled": False,
    "enabled_unreleased": False,
    "unique_id": None,
    "url_server": SENTRY_URL_SERVER,
    "url_coreui": SENTRY_URL_COREUI,
}

IGNORED_EXCEPTIONS = [
    # serial exceptions in octoprint.util.comm
    (
        serial.SerialException,
        lambda exc, logger, plugin, cb: logger == "octoprint.util.comm",
    ),
    # KeyboardInterrupts
    KeyboardInterrupt,
    # IOErrors of any kind due to a full file system
    (
        IOError,
        lambda exc, logger, plugin, cb: exc.errorgetattr(exc, "errno")  # noqa: B009
        and exc.errno in (getattr(errno, "ENOSPC"),),  # noqa: B009
    ),
    # RequestExceptions of any kind
    requests.exceptions.RequestException,
    # Tornado WebSocketErrors of any kind
    tornado.websocket.WebSocketError,
    # Anything triggered by or in third party plugin Astroprint
    (
        Exception,
        lambda exc, logger, plugin, cb: logger.startswith("octoprint.plugins.astroprint")
        or plugin == "astroprint"
        or cb.startswith("octoprint_astroprint."),
    ),
]

try:
    # noinspection PyUnresolvedReferences
    from octoprint.plugins.backup import InsufficientSpace

    # if the backup plugin is enabled, ignore InsufficientSpace errors from it as well
    IGNORED_EXCEPTIONS.append(InsufficientSpace)

    del InsufficientSpace
except ImportError:
    pass


class ErrorTrackingPlugin(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SimpleApiPlugin,
):
    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "name": gettext("Error Tracking"),
                "template": "errortracking_settings.jinja2",
                "custom_bindings": False,
            },
            {"type": "generic", "template": "errortracking_javascripts.jinja2"},
        ]

    def get_template_vars(self):
        enabled = self._settings.get_boolean(["enabled"])
        enabled_unreleased = self._settings.get_boolean(["enabled_unreleased"])

        return {
            "enabled": _is_enabled(enabled, enabled_unreleased),
            "unique_id": self._settings.get(["unique_id"]),
            "url_coreui": self._settings.get(["url_coreui"]),
        }

    def get_assets(self):
        return {"js": ["js/sentry.min.js", "js/errortracking.js"]}

    def get_settings_defaults(self):
        return SETTINGS_DEFAULTS

    def on_settings_save(self, data):
        old_enabled = _is_enabled(
            self._settings.get_boolean(["enabled"]),
            self._settings.get_boolean(["enabled_unreleased"]),
        )

        octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

        enabled = _is_enabled(
            self._settings.get_boolean(["enabled"]),
            self._settings.get_boolean(["enabled_unreleased"]),
        )

        if old_enabled != enabled:
            _enable_errortracking()

    def on_api_get(self, request):
        return jsonify(**self.get_template_vars())


_enabled = False


def _enable_errortracking():
    # this is a bit hackish, but we want to enable error tracking as early in the platform lifecycle as possible
    # and hence can't wait until our implementation is initialized and injected with settings

    from octoprint.settings import settings

    global _enabled

    if _enabled:
        return

    version = get_octoprint_version_string()

    s = settings()
    plugin_defaults = {"plugins": {"errortracking": SETTINGS_DEFAULTS}}

    enabled = s.getBoolean(
        ["plugins", "errortracking", "enabled"], defaults=plugin_defaults
    )
    enabled_unreleased = s.getBoolean(
        ["plugins", "errortracking", "enabled_unreleased"], defaults=plugin_defaults
    )
    url_server = s.get(
        ["plugins", "errortracking", "url_server"], defaults=plugin_defaults
    )
    unique_id = s.get(["plugins", "errortracking", "unique_id"], defaults=plugin_defaults)
    if unique_id is None:
        import uuid

        unique_id = str(uuid.uuid4())
        s.set(
            ["plugins", "errortracking", "unique_id"],
            unique_id,
            defaults=plugin_defaults,
        )
        s.save()

    if _is_enabled(enabled, enabled_unreleased):
        import sentry_sdk

        from octoprint.plugin import plugin_manager

        def _before_send(event, hint):
            if "exc_info" not in hint:
                # we only want exceptions
                return None

            handled = True
            logger = event.get("logger", "")
            plugin = event.get("extra", {}).get("plugin", None)
            callback = event.get("extra", {}).get("callback", None)

            for ignore in IGNORED_EXCEPTIONS:
                if isinstance(ignore, tuple):
                    ignored_exc, matcher = ignore
                else:
                    ignored_exc = ignore
                    matcher = lambda *args: True

                exc = hint["exc_info"][1]
                if isinstance(exc, ignored_exc) and matcher(
                    exc, logger, plugin, callback
                ):
                    # exception ignored for logger, plugin and/or callback
                    return None

                elif isinstance(ignore, type):
                    if isinstance(hint["exc_info"][1], ignore):
                        # exception ignored
                        return None

            if event.get("exception") and event["exception"].get("values"):
                handled = not any(
                    map(
                        lambda x: x.get("mechanism")
                        and not x["mechanism"].get("handled", True),
                        event["exception"]["values"],
                    )
                )

            if handled:
                # error is handled, restrict further based on logger
                if logger != "" and not (
                    logger.startswith("octoprint.") or logger.startswith("tornado.")
                ):
                    # we only want errors logged by loggers octoprint.* or tornado.*
                    return None

                if logger.startswith("octoprint.plugins."):
                    plugin_id = logger.split(".")[2]
                    plugin_info = plugin_manager().get_plugin_info(plugin_id)
                    if plugin_info is None or not plugin_info.bundled:
                        # we only want our active bundled plugins
                        return None

                if plugin is not None:
                    plugin_info = plugin_manager().get_plugin_info(plugin)
                    if plugin_info is None or not plugin_info.bundled:
                        # we only want our active bundled plugins
                        return None

            return event

        sentry_sdk.init(url_server, release=version, before_send=_before_send)

        with sentry_sdk.configure_scope() as scope:
            scope.user = {"id": unique_id}

        logging.getLogger("octoprint.plugins.errortracking").info(
            "Initialized error tracking"
        )
        _enabled = True


def _is_enabled(enabled, enabled_unreleased):
    return enabled and (enabled_unreleased or is_released_octoprint_version())


def __plugin_enable__():
    _enable_errortracking()


__plugin_name__ = "Error Tracking"
__plugin_author__ = "Gina Häußge"
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = ErrorTrackingPlugin()



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import json
import time

import requests
import websocket


def build_base_url(
    https=False, httpuser=None, httppass=None, host=None, port=None, prefix=None
):
    protocol = "https" if https else "http"
    httpauth = f"{httpuser}:{httppass}@" if httpuser and httppass else ""
    host = host if host else "127.0.0.1"
    port = f":{port}" if port else ":5000"
    prefix = prefix if prefix else ""

    return f"{protocol}://{httpauth}{host}{port}{prefix}"


class SocketTimeout(BaseException):
    pass


class SocketClient:
    def __init__(self, url, use_ssl=False, daemon=True, **kwargs):
        self._url = url
        self._use_ssl = use_ssl
        self._daemon = daemon
        self._ws_kwargs = kwargs

        self._seen_open = False
        self._seen_close = False
        self._waiting_for_reconnect = False

        self._ws = None
        self._thread = None

        # hello world

    def _prepare(self):
        """Prepares socket and thread for a new connection."""

        # close the socket if it's currently open
        if self._ws is not None:
            try:
                self._ws.close()
            except Exception:
                # we can't handle that in any meaningful way right now
                pass

        # prepare a bunch of callback methods
        import functools

        callbacks = {}
        for callback in ("on_open", "on_message", "on_error", "on_close"):
            # now, normally we could just use functools.partial for something like
            # this, but websocket does a type check against a python function type
            # which the use of partial makes fail, so we have to use a lambda
            # wrapper here, and since we need to pick the current value of
            # callback from the current scope we need a factory method for that...
            def factory(cb):
                return lambda *fargs, **fkwargs: functools.partial(self._on_callback, cb)(
                    *fargs, **fkwargs
                )

            callbacks[callback] = factory(callback)

        # initialize socket instance with url and callbacks
        kwargs = dict(self._ws_kwargs)
        kwargs.update(callbacks)
        self._ws = websocket.WebSocketApp(self._url, **kwargs)

        # initialize thread
        import threading

        self._thread = threading.Thread(target=self._on_thread_run)
        self._thread.daemon = self._daemon

    def _on_thread_run(self):
        """Has the socket run forever (aka until closed...)."""
        self._ws.run_forever()

    def _on_callback(self, cb, *args, **kwargs):
        """
        Callback for socket events.

        Will call any callback method defined on ``self`` that matches ``cb``
        prefixed with an ``_ws_``, then will call any callback method provided in
        the socket keyword arguments (``self._ws_kwargs``) that matches ``cb``.

        Calling args and kwargs will be the ones passed to ``_on_callback``.

        Arguments:
            cb (str): the callback type
        """
        internal = "_ws_" + cb
        if hasattr(self, internal):
            cb_func = getattr(self, internal)
            if callable(cb_func):
                cb_func(*args, **kwargs)

        cb_func = self._ws_kwargs.get(cb, None)
        if callable(cb_func):
            cb_func(*args, **kwargs)

    def _ws_on_open(self, ws):
        """
        Callback for socket on_open event.

        Used only to track active reconnection attempts.
        """

        if not self._waiting_for_reconnect:
            return
        if ws != self._ws:
            return
        self._seen_open = True

    def _ws_on_close(self, ws):
        """
        Callback for socket on_close event.

        Used only to track active reconnection attempts.
        """

        if not self._waiting_for_reconnect:
            return
        if ws != self._ws:
            return
        self._seen_close = True

    def connect(self):
        """Connects the socket."""
        self._prepare()
        self._thread.start()

    def wait(self, timeout=None):
        """Waits for the closing of the socket or the timeout."""
        start = None

        def test_condition():
            if timeout and start and start + timeout > time.time():
                raise SocketTimeout()

        start = time.time()
        while self._thread.is_alive():
            test_condition()
            self._thread.join(timeout=1.0)

    @property
    def is_connected(self):
        """Whether the web socket is connected or not."""
        return self._thread and self._ws and self._thread.is_alive()

    def disconnect(self):
        """Disconnect the web socket."""
        if self._ws:
            self._ws.close()

    def reconnect(self, timeout=None, disconnect=True):
        """
        Tries to reconnect to the web socket.

        If timeout is set will try to reconnect over the specified timeout in seconds
        and return False if the connection could not be re-established.

        If no timeout is provided, the method will block until the connection could
        be re-established.

        If disconnect is set to ``True`` will disconnect the socket explicitly
        first if it is currently connected.

        Arguments:
            timeout (number): timeout in seconds to wait for the reconnect to happen.
            disconnect (bool): Whether to disconnect explicitly from the socket if
               a connection is currently established (True, default) or not (False).

        Returns:
            bool - True if the reconnect was successful, False otherwise.
        """

        self._seen_close = False
        self._seen_open = False
        self._waiting_for_reconnect = True

        if not self.is_connected:
            # not connected, so we are already closed
            self._seen_close = True
        elif disconnect:
            # connected and disconnect is True, so we disconnect
            self.disconnect()

        start = None
        if timeout:
            timeout_condition = (
                lambda: start is not None and time.time() > start + timeout
            )
        else:
            timeout_condition = lambda: False

        start = time.time()
        while not timeout_condition():
            if self._seen_close and self._seen_open:
                # we saw a connection close and open, so a reconnect, success!
                return True
            else:
                # try to connect
                self.connect()

            # sleep a bit
            time.sleep(1.0)

        # if we land here the timeout condition became True without us seeing
        # a reconnect, that's a failure
        return False


class Client:
    def __init__(self, baseurl, apikey):
        self.baseurl = baseurl
        self.apikey = apikey

    def prepare_request(self, method=None, path=None, params=None):
        url = None
        if self.baseurl:
            while path.startswith("/"):
                path = path[1:]
            url = self.baseurl + "/" + path
        return requests.Request(
            method=method, url=url, params=params, headers={"X-Api-Key": self.apikey}
        ).prepare()

    def request(
        self,
        method,
        path,
        data=None,
        files=None,
        encoding=None,
        params=None,
        timeout=None,
    ):
        if timeout is None:
            timeout = 30

        s = requests.Session()
        request = self.prepare_request(method, path, params=params)
        if data or files:
            if encoding == "json":
                request.prepare_body(None, None, json=data)
            else:
                request.prepare_body(data, files=files)
        response = s.send(request, timeout=timeout)
        return response

    def get(self, path, params=None, timeout=None):
        return self.request("GET", path, params=params, timeout=timeout)

    def post(self, path, data, encoding=None, params=None, timeout=None):
        return self.request(
            "POST", path, data=data, encoding=encoding, params=params, timeout=timeout
        )

    def post_json(self, path, data, params=None, timeout=None):
        return self.post(path, data, encoding="json", params=params, timeout=timeout)

    def post_command(self, path, command, additional=None, timeout=None):
        data = {"command": command}
        if additional:
            data.update(additional)
        return self.post_json(path, data, params=data, timeout=timeout)

    def upload(
        self,
        path,
        file_path,
        additional=None,
        file_name=None,
        content_type=None,
        params=None,
        timeout=None,
    ):
        import os

        if not os.path.isfile(file_path):
            raise ValueError(f"{file_path} cannot be uploaded since it is not a file")

        if file_name is None:
            file_name = os.path.basename(file_path)

        with open(file_path, "rb") as fp:
            if content_type:
                files = {"file": (file_name, fp, content_type)}
            else:
                files = {"file": (file_name, fp)}

            response = self.request(
                "POST", path, data=additional, files=files, params=params, timeout=timeout
            )

        return response

    def delete(self, path, params=None, timeout=None):
        return self.request("DELETE", path, params=params, timeout=timeout)

    def patch(self, path, data, encoding=None, params=None, timeout=None):
        return self.request(
            "PATCH", path, data=data, encoding=encoding, params=params, timeout=timeout
        )

    def put(self, path, data, encoding=None, params=None, timeout=None):
        return self.request(
            "PUT", path, data=data, encoding=encoding, params=params, timeout=timeout
        )

    def create_socket(self, **kwargs):
        import random
        import uuid

        # creates websocket URL for SockJS according to
        # - http://sockjs.github.io/sockjs-protocol/sockjs-protocol-0.3.3.html#section-37
        # - http://sockjs.github.io/sockjs-protocol/sockjs-protocol-0.3.3.html#section-50
        url = "ws://{}/sockjs/{:0>3d}/{}/websocket".format(
            self.baseurl[
                self.baseurl.find("//") + 2 :
            ],  # host + port + prefix, but no protocol
            random.randrange(0, stop=999),  # server_id
            uuid.uuid4(),  # session_id
        )
        use_ssl = self.baseurl.startswith("https:")

        on_open_cb = kwargs.get("on_open", None)
        on_heartbeat_cb = kwargs.get("on_heartbeat", None)
        on_message_cb = kwargs.get("on_message", None)
        on_close_cb = kwargs.get("on_close", None)
        on_error_cb = kwargs.get("on_error", None)
        on_sent_cb = kwargs.get("on_sent", None)
        daemon = kwargs.get("daemon", True)

        def send(ws, data):
            payload = '["' + json.dumps(data).replace('"', '\\"') + '"]'
            ws.send(payload)
            if callable(on_sent_cb):
                on_sent_cb(ws, data)

        def authenticate(ws):
            # perform passive login to retrieve username and session key for API key
            response = self.post("/api/login", {"passive": True})
            response.raise_for_status()
            data = response.json()

            # prepare auth payload
            auth_message = {"auth": "{name}:{session}".format(**data)}

            # send it
            send(ws, auth_message)

        def on_message(ws, message):
            message_type = message[0]

            if message_type == "h":
                # "heartbeat" message
                if callable(on_heartbeat_cb):
                    on_heartbeat_cb(ws)
                    return
            elif message_type == "o":
                # "open" message
                return
            elif message_type == "c":
                # "close" message
                return

            if not callable(on_message_cb):
                return

            message_body = message[1:]
            if not message_body:
                return

            data = json.loads(message_body)

            if message_type == "m":
                data = [
                    data,
                ]

            for d in data:
                for internal_type, internal_message in d.items():
                    on_message_cb(ws, internal_type, internal_message)
                    if internal_type == "connected":
                        # we just got connected to the server, authenticate
                        authenticate(ws)

        def on_open(ws):
            if callable(on_open_cb):
                on_open_cb(ws)

        def on_close(ws):
            if callable(on_close_cb):
                on_close_cb(ws)

        def on_error(ws, error):
            if callable(on_error_cb):
                on_error_cb(ws, error)

        class CustomSocketClient(SocketClient):
            def auth(self, username, key):
                self.send({"auth": f"{username}:{key}"})

            def throttle(self, factor):
                self.send({"throttle": factor})

            def send(self, data):
                send(self._ws, data)

        socket = CustomSocketClient(
            url,
            use_ssl=use_ssl,
            daemon=daemon,
            on_open=on_open,
            on_message=on_message,
            on_close=on_close,
            on_error=on_error,
        )
        socket.connect()

        return socket



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class FeatureConfig(BaseModel):
    temperatureGraph: bool = True
    """Whether to enable the temperature graph in the UI or not."""

    sdSupport: bool = True
    """Specifies whether support for SD printing and file management should be enabled."""

    keyboardControl: bool = True
    """Whether to enable the keyboard control feature in the control tab."""

    pollWatched: bool = False
    """Whether to actively poll the watched folder (true) or to rely on the OS's file system notifications instead (false)."""

    modelSizeDetection: bool = True
    """Whether to enable model size detection and warning (true) or not (false)."""

    rememberFileFolder: bool = False
    """Whether to remember the selected folder on the file manager."""

    printStartConfirmation: bool = False
    """Whether to show a confirmation on print start (true) or not (false)"""

    printCancelConfirmation: bool = True
    """Whether to show a confirmation on print cancelling (true) or not (false)"""

    uploadOverwriteConfirmation: bool = True

    autoUppercaseBlacklist: List[str] = ["M117", "M118"]
    """Commands that should never be auto-uppercased when sent to the printer through the Terminal tab."""

    g90InfluencesExtruder: bool = False
    """Whether `G90`/`G91` also influence absolute/relative mode of extruders."""

    enforceReallyUniversalFilenames: bool = False
    """Replace all special characters and spaces with text equivalent to make them universally compatible. Most OS filesystems work fine with unicode characters, but just in case you can revert to the older behaviour by setting this to true."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import inspect
import sys

from flask_babel import gettext

import octoprint.plugin
from octoprint.access import ADMIN_GROUP, USER_GROUP


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class ServerCommandsSubwizard:
    def _is_servercommands_wizard_firstrunonly(self):
        return True

    def _is_servercommands_wizard_required(self):
        system_shutdown_command = self._settings.global_get(
            ["server", "commands", "systemShutdownCommand"]
        )
        system_restart_command = self._settings.global_get(
            ["server", "commands", "systemRestartCommand"]
        )
        server_restart_command = self._settings.global_get(
            ["server", "commands", "serverRestartCommand"]
        )

        return not (
            system_shutdown_command and system_restart_command and server_restart_command
        )

    def _get_servercommands_wizard_details(self):
        return {"required": self._is_servercommands_wizard_required()}

    def _get_servercommands_wizard_name(self):
        return gettext("Server Commands")


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class AclSubwizard:
    def _is_acl_wizard_firstrunonly(self):
        return False

    def _is_acl_wizard_required(self):
        return not self._user_manager.has_been_customized()

    def _get_acl_wizard_details(self):
        return {"required": self._is_acl_wizard_required()}

    def _get_acl_wizard_name(self):
        return gettext("Access Control")

    def _get_acl_additional_wizard_template_data(self):
        return {"mandatory": self._is_acl_wizard_required()}

    @octoprint.plugin.BlueprintPlugin.route("/acl", methods=["POST"])
    def acl_wizard_api(self):
        from flask import abort, request

        from octoprint.server.api import NO_CONTENT

        if (
            not self._settings.global_get(["server", "firstRun"])
            and self._user_manager.has_been_customized()
        ):
            abort(404)

        data = request.get_json(silent=True)
        if data is None:
            data = request.values

        if (
            "user" in data
            and "pass1" in data
            and "pass2" in data
            and data["pass1"] == data["pass2"]
        ):
            # configure access control
            self._user_manager.add_user(
                data["user"],
                data["pass1"],
                True,
                [],
                [USER_GROUP, ADMIN_GROUP],
                overwrite=True,
            )
        self._settings.save()
        return NO_CONTENT


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class OnlineCheckSubwizard:
    def _is_onlinecheck_wizard_firstrunonly(self):
        return False

    def _is_onlinecheck_wizard_required(self):
        return self._settings.global_get(["server", "onlineCheck", "enabled"]) is None

    def _get_onlinecheck_wizard_details(self):
        return {"required": self._is_onlinecheck_wizard_required()}

    def _get_onlinecheck_wizard_name(self):
        return gettext("Online Connectivity Check")

    def _get_onlinecheck_additional_wizard_template_data(self):
        return {"mandatory": self._is_onlinecheck_wizard_required()}


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class PluginBlacklistSubwizard:
    def _is_pluginblacklist_wizard_firstrunonly(self):
        return False

    def _is_pluginblacklist_wizard_required(self):
        return self._settings.global_get(["server", "pluginBlacklist", "enabled"]) is None

    def _get_pluginblacklist_wizard_details(self):
        return {"required": self._is_pluginblacklist_wizard_required()}

    def _get_pluginblacklist_wizard_name(self):
        return gettext("Plugin Blacklist")

    def _get_pluginblacklist_additional_wizard_template_data(self):
        return {"mandatory": self._is_pluginblacklist_wizard_required()}


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class PrinterProfileSubwizard:
    def _is_printerprofile_wizard_firstrunonly(self):
        return True

    def _is_printerprofile_wizard_required(self):
        return (
            self._printer_profile_manager.is_default_unmodified()
            and self._printer_profile_manager.profile_count == 1
        )

    def _get_printerprofile_wizard_details(self):
        return {"required": self._is_printerprofile_wizard_required()}

    def _get_printerprofile_wizard_name(self):
        return gettext("Default Printer Profile")


Subwizards = type(
    "Subwizards",
    tuple(
        cls
        for clsname, cls in inspect.getmembers(sys.modules[__name__], inspect.isclass)
        if clsname.endswith("Subwizard")
    ),
    {},
)



# Based on pydantic-settings version 0.2.5 by Daniel Daniels, licensed under MIT
#
# https://github.com/danields761/pydantic-settings

from typing import Type

from class_doc import extract_docs_from_cls_obj
from pydantic import BaseModel


def apply_attributes_docs(
    model: Type[BaseModel], override_existing: bool = True
) -> None:
    """
    Apply model attributes documentation in-place. Resulted docs are placed
    inside :code:`field.schema.description` for *pydantic* model field.
    :param model: any pydantic model
    :param override_existing: override existing descriptions
    """
    docs = extract_docs_from_cls_obj(model)

    for field in model.__fields__.values():
        if field.field_info.description and not override_existing:
            continue

        try:
            field.field_info.description = '\n'.join(docs[field.name])
        except KeyError:
            pass


def with_attrs_docs(
    model_cls: Type[BaseModel]
) -> Type[BaseModel]:
    """
    Applies :py:func:`.apply_attributes_docs`.
    """

    def decorator(maybe_model_cls: Type[BaseModel]) -> Type[BaseModel]:
        apply_attributes_docs(
            maybe_model_cls
        )
        return maybe_model_cls

    return decorator(model_cls)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import List, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class SubscriptionTypeEnum(str, Enum):
    system = "system"
    gcode = "gcode"


@with_attrs_docs
class EventSubscription(BaseModel):
    event: str
    """The event to subscribe to."""

    name: Optional[str] = None
    """The event name to show on the UI"""

    command: str
    """The command to execute when the event is triggered, either a GCODE or a system command."""

    type: SubscriptionTypeEnum
    """The type of the command."""

    enabled: bool = True
    """Whether the event subscription should be enabled."""

    debug: bool = False
    """If set to `true`, OctoPrint will log the command after performing all placeholder replacements."""


@with_attrs_docs
class EventsConfig(BaseModel):
    enabled: bool = True
    """Whether event subscriptions should be enabled or not."""

    subscriptions: List[EventSubscription] = []
    """A list of event subscriptions."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import copy
import logging
import os
import sys
import threading

import psutil

from octoprint.plugin import EnvironmentDetectionPlugin
from octoprint.util import yaml
from octoprint.util.platform import get_os
from octoprint.util.version import get_python_version_string


class EnvironmentDetector:
    def __init__(self, plugin_manager):
        self._plugin_manager = plugin_manager

        self._cache = None
        self._cache_lock = threading.RLock()

        self._logger = logging.getLogger(__name__)

        try:
            self._environment_plugins = self._plugin_manager.get_implementations(
                EnvironmentDetectionPlugin
            )
        except Exception:
            # just in case, see #3100...
            self._logger.exception(
                "There was an error fetching EnvironmentDetectionPlugins from the plugin manager"
            )
            self._environment_plugins = []

    @property
    def environment(self):
        with self._cache_lock:
            if self._cache is None:
                self.run_detection()
            return copy.deepcopy(self._cache)

    def run_detection(self, notify_plugins=True):
        try:
            environment = {}
            environment["os"] = self._detect_os()
            environment["python"] = self._detect_python()
            environment["hardware"] = self._detect_hardware()

            plugin_result = self._detect_from_plugins()
            if plugin_result:
                environment["plugins"] = plugin_result

            with self._cache_lock:
                self._cache = environment

            if notify_plugins:
                self.notify_plugins()

            return environment
        except Exception:
            self._logger.exception("Unexpected error while detecting environment")
            with self._cache_lock:
                self._cache = {}
                return self._cache

    def _detect_os(self):
        return {
            "id": get_os(),
            "platform": sys.platform,
            "bits": 64 if sys.maxsize > 2**32 else 32,
        }

    def _detect_python(self):
        result = {"version": "unknown", "pip": "unknown"}

        # determine python version
        try:
            result["version"] = get_python_version_string()
        except Exception:
            self._logger.exception("Error detecting python version")

        # determine if we are running from a virtual environment
        try:
            if hasattr(sys, "real_prefix") or (
                hasattr(sys, "base_prefix")
                and os.path.realpath(sys.prefix) != os.path.realpath(sys.base_prefix)
            ):
                result["virtualenv"] = sys.prefix
        except Exception:
            self._logger.exception(
                "Error detecting whether we are running in a virtual environment"
            )

        # try to find pip version
        try:
            import pkg_resources

            result["pip"] = pkg_resources.get_distribution("pip").version
        except Exception:
            self._logger.exception("Error detecting pip version")

        return result

    def _detect_hardware(self):
        result = {"cores": "unknown", "freq": "unknown", "ram": "unknown"}

        try:
            cores = psutil.cpu_count()
            cpu_freq = psutil.cpu_freq()
            ram = psutil.virtual_memory()
            if cores:
                result["cores"] = cores
            if cpu_freq and hasattr(cpu_freq, "max"):
                result["freq"] = cpu_freq.max
            if ram and hasattr(ram, "total"):
                result["ram"] = ram.total
        except Exception:
            self._logger.exception("Error while detecting hardware environment")

        return result

    def _detect_from_plugins(self):
        result = {}

        for implementation in self._environment_plugins:
            try:
                additional = implementation.get_additional_environment()
                if (
                    additional is not None
                    and isinstance(additional, dict)
                    and len(additional)
                ):
                    result[implementation._identifier] = additional
            except Exception:
                self._logger.exception(
                    "Error while fetching additional "
                    "environment data from plugin {}".format(implementation._identifier),
                    extra={"plugin": implementation._identifier},
                )

        return result

    def log_detected_environment(self, only_to_handler=None):
        def _log(message, level=logging.INFO):
            if only_to_handler is not None:
                import octoprint.logging

                octoprint.logging.log_to_handler(
                    self._logger, only_to_handler, level, message, []
                )
            else:
                self._logger.log(level, message)

        try:
            _log(self._format())
        except Exception:
            self._logger.exception("Error logging detected environment")

    def _format(self):
        with self._cache_lock:
            if self._cache is None:
                self.run_detection()
            environment = copy.deepcopy(self._cache)

        dumped_environment = yaml.dump(environment, pretty=True).strip()
        environment_lines = "\n".join(
            map(lambda x: f"|  {x}", dumped_environment.split("\n"))
        )
        return "Detected environment is Python {} under {} ({}). Details:\n{}".format(
            environment["python"]["version"],
            environment["os"]["id"].title(),
            environment["os"]["platform"],
            environment_lines,
        )

    def notify_plugins(self):
        with self._cache_lock:
            if self._cache is None:
                self.run_detection(notify_plugins=False)
            environment = copy.deepcopy(self._cache)

        for implementation in self._environment_plugins:
            try:
                implementation.on_environment_detected(environment)
            except Exception:
                self._logger.exception(
                    "Error while sending environment "
                    "detection result to plugin {}".format(implementation._identifier)
                )



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

import requests

from ..exceptions import CannotCheckOffline, ConfigurationInvalid, NetworkError

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.etag")


def get_latest(target, check, online=True, *args, **kwargs):
    if not online:
        raise CannotCheckOffline()

    url = check.get("header_url", check.get("url"))
    header = check.get("header_name")

    if url is None or header is None:
        raise ConfigurationInvalid(
            "HTTP header version check needs header_url or url and header_name set"
        )

    current = check.get("current")
    method = check.get("header_method", "head")
    prefix = check.get("header_prefix", header)
    if prefix:
        prefix = f"{prefix} "

    try:
        with requests.request(method, url) as r:
            latest = r.headers.get(header)
    except Exception as exc:
        raise NetworkError(cause=exc)

    information = {
        "local": {
            "name": "{}{}".format(prefix, current if current else "-"),
            "value": current,
        },
        "remote": {
            "name": "{}{}".format(prefix, latest if latest else "-"),
            "value": latest,
        },
    }

    logger.debug(
        "Target: {}, local: {}, remote: {}".format(
            target, information["local"]["name"], information["remote"]["name"]
        )
    )

    return information, current is None or current == latest or latest is None



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import re
import threading
import time

import wrapt

import octoprint.access.users
import octoprint.events
import octoprint.plugin
import octoprint.printer
import octoprint.server
import octoprint.timelapse
import octoprint.vendor.sockjs.tornado
import octoprint.vendor.sockjs.tornado.proto
import octoprint.vendor.sockjs.tornado.session
import octoprint.vendor.sockjs.tornado.util
from octoprint.access.groups import GroupChangeListener
from octoprint.access.permissions import Permissions
from octoprint.access.users import LoginStatusListener, SessionUser
from octoprint.events import Events
from octoprint.settings import settings
from octoprint.util import RepeatedTimer
from octoprint.util.json import dumps as json_dumps
from octoprint.util.version import get_python_version_string


class ThreadSafeSession(octoprint.vendor.sockjs.tornado.session.Session):
    def __init__(self, conn, server, session_id, expiry=None):
        octoprint.vendor.sockjs.tornado.session.Session.__init__(
            self, conn, server, session_id, expiry=expiry
        )

    def set_handler(self, handler, start_heartbeat=True):
        if getattr(handler, "__orig_send_pack", None) is None:
            orig_send_pack = handler.send_pack
            mutex = threading.RLock()

            def send_pack(*args, **kwargs):
                with mutex:
                    return orig_send_pack(*args, **kwargs)

            handler.send_pack = send_pack
            handler.__orig_send_pack = orig_send_pack

        return octoprint.vendor.sockjs.tornado.session.Session.set_handler(
            self, handler, start_heartbeat=start_heartbeat
        )

    def remove_handler(self, handler):
        result = octoprint.vendor.sockjs.tornado.session.Session.remove_handler(
            self, handler
        )

        if getattr(handler, "__orig_send_pack", None) is not None:
            handler.send_pack = handler.__orig_send_pack
            delattr(handler, "__orig_send_pack")

        return result


class JsonEncodingSessionWrapper(wrapt.ObjectProxy):
    def send_message(self, msg, stats=True, binary=False):
        self.send_jsonified(
            json_dumps(octoprint.vendor.sockjs.tornado.util.bytes_to_str(msg)),
            stats,
        )


class PrinterStateConnection(
    octoprint.vendor.sockjs.tornado.SockJSConnection,
    octoprint.printer.PrinterCallback,
    LoginStatusListener,
    GroupChangeListener,
):
    _event_permissions = {
        Events.USER_LOGGED_IN: [Permissions.ADMIN],
        Events.USER_LOGGED_OUT: [Permissions.ADMIN],
        "*": [],
    }

    _event_payload_processors = {
        Events.CLIENT_OPENED: [
            lambda user, payload: payload
            if user.has_permission(Permissions.ADMIN)
            else {}
        ],
        Events.CLIENT_AUTHED: [
            lambda user, payload: payload
            if user.has_permission(Permissions.ADMIN)
            else {}
        ],
        "*": [],
    }

    # TODO: Permissions should be overridable from plugins, this special case stuff here is a hack
    _emit_permissions = {
        "connected": [],
        "reauthRequired": [],
        "plugin": lambda payload: []
        if payload.get("plugin") in ("backup", "softwareupdate")
        and settings().getBoolean(["server", "firstRun"])
        else [Permissions.STATUS],
        "*": [Permissions.STATUS],
    }

    _unauthed_backlog_max = 100

    def __init__(
        self,
        printer,
        fileManager,
        analysisQueue,
        userManager,
        groupManager,
        eventManager,
        pluginManager,
        connectivityChecker,
        session,
    ):
        if isinstance(session, octoprint.vendor.sockjs.tornado.session.Session):
            session = JsonEncodingSessionWrapper(session)

        octoprint.vendor.sockjs.tornado.SockJSConnection.__init__(self, session)

        self._logger = logging.getLogger(__name__)

        self._temperatureBacklog = []
        self._temperatureBacklogMutex = threading.Lock()
        self._logBacklog = []
        self._logBacklogMutex = threading.Lock()
        self._messageBacklog = []
        self._messageBacklogMutex = threading.Lock()

        self._unauthed_backlog = []
        self._unauthed_backlog_mutex = threading.RLock()

        self._printer = printer
        self._fileManager = fileManager
        self._analysisQueue = analysisQueue
        self._userManager = userManager
        self._groupManager = groupManager
        self._eventManager = eventManager
        self._pluginManager = pluginManager
        self._connectivityChecker = connectivityChecker

        self._remoteAddress = None
        self._user = self._userManager.anonymous_user_factory()

        self._throttle_factor = 1
        self._last_current = 0
        self._base_rate_limit = 0.5

        self._held_back_current = None
        self._held_back_mutex = threading.RLock()

        self._register_hooks = self._pluginManager.get_hooks(
            "octoprint.server.sockjs.register"
        )
        self._authed_hooks = self._pluginManager.get_hooks(
            "octoprint.server.sockjs.authed"
        )
        self._emit_hooks = self._pluginManager.get_hooks("octoprint.server.sockjs.emit")

        self._registered = False
        self._authed = False
        self._initial_data_sent = False

        self._subscriptions_active = False
        self._subscriptions = {"state": False, "plugins": [], "events": []}

        self._keep_alive = RepeatedTimer(
            60, self._keep_alive_callback, condition=lambda: self._authed
        )

    @staticmethod
    def _get_remote_address(info):
        forwarded_for = info.headers.get("X-Forwarded-For")
        if forwarded_for is not None:
            return forwarded_for.split(",")[0]
        return info.ip

    def _keep_alive_callback(self):
        if not self._authed:
            return
        if not isinstance(self._user, SessionUser):
            return
        self._user.touch()

    def __str__(self):
        if self._remoteAddress:
            return f"{self!r} connected to {self._remoteAddress}"
        else:
            return f"Unconnected {self!r}"

    def on_open(self, info):
        self._pluginManager.register_message_receiver(self.on_plugin_message)
        self._remoteAddress = self._get_remote_address(info)
        self._logger.info("New connection from client: %s" % self._remoteAddress)

        self._userManager.register_login_status_listener(self)
        self._groupManager.register_listener(self)

        plugin_signature = lambda impl: "{}:{}".format(
            impl._identifier, impl._plugin_version
        )
        template_plugins = list(
            map(
                plugin_signature,
                self._pluginManager.get_implementations(octoprint.plugin.TemplatePlugin),
            )
        )
        asset_plugins = list(
            map(
                plugin_signature,
                self._pluginManager.get_implementations(octoprint.plugin.AssetPlugin),
            )
        )
        ui_plugins = sorted(set(template_plugins + asset_plugins))

        import hashlib

        plugin_hash = hashlib.md5()
        plugin_hash.update(",".join(ui_plugins).encode("utf-8"))

        config_hash = settings().config_hash

        # connected => update the API key, might be necessary if the client was left open while the server restarted
        self._emit(
            "connected",
            {
                "version": octoprint.server.VERSION,
                "display_version": octoprint.server.DISPLAY_VERSION,
                "branch": octoprint.server.BRANCH,
                "python_version": get_python_version_string(),
                "plugin_hash": plugin_hash.hexdigest(),
                "config_hash": config_hash,
                "debug": octoprint.server.debug,
                "safe_mode": octoprint.server.safe_mode,
                "online": self._connectivityChecker.online,
                "permissions": [permission.as_dict() for permission in Permissions.all()],
            },
        )

        self._eventManager.fire(
            Events.CLIENT_OPENED, {"remoteAddress": self._remoteAddress}
        )
        self._register()

    def on_close(self):
        self._user = self._userManager.anonymous_user_factory()
        self._groupManager.unregister_listener(self)
        self._userManager.unregister_login_status_listener(self)

        self._unregister()
        self._eventManager.fire(
            Events.CLIENT_CLOSED, {"remoteAddress": self._remoteAddress}
        )

        self._logger.info("Client connection closed: %s" % self._remoteAddress)

        self._on_logout()
        self._remoteAddress = None
        self._pluginManager.unregister_message_receiver(self.on_plugin_message)

    def on_message(self, message):
        try:
            import json

            message = json.loads(message)
        except Exception:
            self._logger.warning(
                "Invalid JSON received from client {}, ignoring: {!r}".format(
                    self._remoteAddress, message
                )
            )
            return

        if "auth" in message:
            try:
                parts = message["auth"].split(":")
                if not len(parts) == 2:
                    raise ValueError()
            except ValueError:
                self._logger.warning(
                    "Got invalid auth message from client {}, ignoring: {!r}".format(
                        self._remoteAddress, message["auth"]
                    )
                )
            else:
                user_id, user_session = parts

                if self._userManager.validate_user_session(user_id, user_session):
                    user = self._userManager.find_user(
                        userid=user_id, session=user_session
                    )
                    self._on_login(user)
                else:
                    self._logger.warning(
                        f"Unknown user/session combo: {user_id}:{user_session}"
                    )
                    self._on_logout()
                    self._sendReauthRequired("stale")

            self._register()

        elif "throttle" in message:
            try:
                throttle = int(message["throttle"])
                if throttle < 1:
                    raise ValueError()
            except ValueError:
                self._logger.warning(
                    "Got invalid throttle factor from client {}, ignoring: {!r}".format(
                        self._remoteAddress, message["throttle"]
                    )
                )
            else:
                self._throttle_factor = throttle
                self._logger.debug(
                    "Set throttle factor for client {} to {}".format(
                        self._remoteAddress, self._throttle_factor
                    )
                )

        elif "subscribe" in message:
            if not self._subscriptions_active:
                self._subscriptions_active = True
                self._logger.debug("Client makes use of subscriptions")

            def list_or_boolean(value):
                if isinstance(value, list):
                    return value
                elif isinstance(value, bool):
                    return [] if not value else None
                else:
                    raise ValueError("value must be a list or boolean")

            def regex_or_boolean(value):
                if isinstance(value, str):
                    try:
                        return re.compile(value)
                    except Exception:
                        raise ValueError("value must be a valid regex")
                elif isinstance(value, bool):
                    return value
                else:
                    raise ValueError("value must be a string or boolean")

            try:
                subscribe = message["subscribe"]

                state = subscribe.get("state", False)
                if isinstance(state, bool):
                    if state:
                        state = {"logs": True, "messages": False}
                elif isinstance(state, dict):
                    logs = regex_or_boolean(state.get("logs", False))
                    messages = regex_or_boolean(state.get("messages", False))
                    state = {
                        "logs": logs,
                        "messages": messages,
                    }

                plugins = list_or_boolean(subscribe.get("plugins", []))
                events = list_or_boolean(subscribe.get("events", []))

            except ValueError as e:
                self._logger.warning(
                    "Got invalid subscription message from client {}, ignoring: {!r} ({}) ".format(
                        self._remoteAddress, message["subscribe"], str(e)
                    )
                )
            else:
                old_state = self._subscriptions["state"]
                self._subscriptions["state"] = state
                self._subscriptions["plugins"] = plugins
                self._subscriptions["events"] = events

                if state and not old_state:
                    # trigger initial data
                    self._printer.send_initial_callback(self)
                elif old_state and not state:
                    self._initial_data_sent = False

    def on_printer_send_current_data(self, data):
        if not self._user.has_permission(Permissions.STATUS):
            return

        if self._subscriptions_active and not self._subscriptions["state"]:
            return

        if not self._initial_data_sent:
            self._logger.debug("Initial data not yet send, dropping current message")
            return

        # make sure we rate limit the updates according to our throttle factor
        with self._held_back_mutex:
            if self._held_back_current is not None:
                self._held_back_current.cancel()
                self._held_back_current = None

            now = time.time()
            delta = (
                self._last_current + self._base_rate_limit * self._throttle_factor - now
            )
            if delta > 0:
                self._held_back_current = threading.Timer(
                    delta, lambda: self.on_printer_send_current_data(data)
                )
                self._held_back_current.start()
                return

        self._last_current = now

        # add current temperature, log and message backlogs to sent data
        with self._temperatureBacklogMutex:
            temperatures = self._temperatureBacklog
            self._temperatureBacklog = []

        with self._logBacklogMutex:
            logs = self._filter_logs(self._logBacklog)
            self._logBacklog = []

        with self._messageBacklogMutex:
            messages = self._filter_messages(self._messageBacklog)
            self._messageBacklog = []

        busy_files = [
            {"origin": v[0], "path": v[1]} for v in self._fileManager.get_busy_files()
        ]
        if (
            "job" in data
            and data["job"] is not None
            and "file" in data["job"]
            and "path" in data["job"]["file"]
            and "origin" in data["job"]["file"]
            and data["job"]["file"]["path"] is not None
            and data["job"]["file"]["origin"] is not None
            and (self._printer.is_printing() or self._printer.is_paused())
        ):
            busy_files.append(
                {
                    "origin": data["job"]["file"]["origin"],
                    "path": data["job"]["file"]["path"],
                }
            )

        data.update(
            {
                "serverTime": time.time(),
                "temps": temperatures,
                "busyFiles": busy_files,
                "markings": list(self._printer.get_markings()),
            }
        )
        if self._user.has_permission(Permissions.MONITOR_TERMINAL):
            data.update(
                {
                    "logs": self._filter_logs(logs),
                    "messages": messages,
                }
            )
        self._emit("current", payload=data)

    def on_printer_send_initial_data(self, data):
        self._initial_data_sent = True
        if self._subscriptions_active and not self._subscriptions["state"]:
            self._logger.debug("Not subscribed to state, dropping history")
            return

        data_to_send = dict(data)

        data_to_send["serverTime"] = time.time()
        if self._user.has_permission(Permissions.MONITOR_TERMINAL):
            data_to_send["logs"] = self._filter_logs(data_to_send.get("logs", []))
            data_to_send["messages"] = self._filter_messages(
                data_to_send.get("messages", [])
            )
        self._emit("history", payload=data_to_send)

    def _filter_state_subscription(self, sub, values):
        if not self._subscriptions_active or self._subscriptions["state"][sub] is True:
            return values

        if self._subscriptions["state"][sub] is False:
            return []

        return [line for line in values if self._subscriptions["state"][sub].search(line)]

    def _filter_logs(self, logs):
        return self._filter_state_subscription("logs", logs)

    def _filter_messages(self, messages):
        return self._filter_state_subscription("messages", messages)

    def sendEvent(self, type, payload=None):
        permissions = self._event_permissions.get(type, self._event_permissions["*"])
        permissions = [x(self._user) if callable(x) else x for x in permissions]
        if not self._user or not all(
            map(lambda p: self._user.has_permission(p), permissions)
        ):
            return

        processors = self._event_payload_processors.get(
            type, self._event_payload_processors["*"]
        )
        for processor in processors:
            payload = processor(self._user, payload)

        self._emit("event", payload={"type": type, "payload": payload})

    def sendTimelapseConfig(self, timelapseConfig):
        self._emit("timelapse", payload=timelapseConfig)

    def sendSlicingProgress(
        self, slicer, source_location, source_path, dest_location, dest_path, progress
    ):
        self._emit(
            "slicingProgress",
            payload={
                "slicer": slicer,
                "source_location": source_location,
                "source_path": source_path,
                "dest_location": dest_location,
                "dest_path": dest_path,
                "progress": progress,
            },
        )

    def sendRenderProgress(self, progress):
        self._emit("renderProgress", {"progress": progress})

    def on_plugin_message(self, plugin, data, permissions=None):
        if (
            self._subscriptions_active
            and self._subscriptions["plugins"] is not None
            and plugin not in self._subscriptions["plugins"]
        ):
            return

        self._emit(
            "plugin", payload={"plugin": plugin, "data": data}, permissions=permissions
        )

    def on_printer_add_log(self, data):
        with self._logBacklogMutex:
            self._logBacklog.append(data)

    def on_printer_add_message(self, data):
        with self._messageBacklogMutex:
            self._messageBacklog.append(data)

    def on_printer_add_temperature(self, data):
        with self._temperatureBacklogMutex:
            self._temperatureBacklog.append(data)

    def on_user_logged_out(self, user, stale=False):
        if (
            user.get_id() == self._user.get_id()
            and hasattr(user, "session")
            and hasattr(self._user, "session")
            and user.session == self._user.session
        ):
            self._logger.info(f"User {user.get_id()} logged out, logging out on socket")
            self._on_logout()

            if stale:
                self._sendReauthRequired("stale")
            else:
                self._sendReauthRequired("logout")

    def on_user_modified(self, user):
        if user.get_id() == self._user.get_id():
            self._sendReauthRequired("modified")

    def on_user_removed(self, userid):
        if self._user.get_id() == userid:
            self._logger.info(f"User {userid} deleted, logging out on socket")
            self._on_logout()
            self._sendReauthRequired("removed")

    def on_group_permissions_changed(self, group, added=None, removed=None):
        if self._user.is_anonymous and group == self._groupManager.guest_group:
            self._sendReauthRequired("modified")

    def on_group_subgroups_changed(self, group, added=None, removed=None):
        if self._user.is_anonymous and group == self._groupManager.guest_group:
            self._sendReauthRequired("modified")

    def _onEvent(self, event, payload):
        if (
            self._subscriptions_active
            and self._subscriptions["events"] is not None
            and event not in self._subscriptions["events"]
        ):
            return

        self.sendEvent(event, payload)

    def _register(self):
        """Register this socket with the system if STATUS permission is available."""

        proceed = True
        for name, hook in self._register_hooks.items():
            try:
                proceed = proceed and hook(self, self._user)
            except Exception:
                self._logger.exception(
                    f"Error processing register hook handler for plugin {name}",
                    extra={"plugin": name},
                )

        if not proceed:
            return

        if self._registered:
            return

        if not self._user.has_permission(Permissions.STATUS):
            return

        # printer
        self._printer.register_callback(self)
        self._printer.send_initial_callback(self)

        # files
        self._fileManager.register_slicingprogress_callback(self)

        # events
        for event in octoprint.events.all_events():
            self._eventManager.subscribe(event, self._onEvent)

        # timelapse
        octoprint.timelapse.register_callback(self)
        octoprint.timelapse.notify_callback(self, timelapse=octoprint.timelapse.current)
        if octoprint.timelapse.current_render_job is not None:
            # This is a horrible hack for now to allow displaying a notification that a render job is still
            # active in the backend on a fresh connect of a client. This needs to be substituted with a proper
            # job management for timelapse rendering, analysis stuff etc that also gets cancelled when prints
            # start and so on.
            #
            # For now this is the easiest way though to at least inform the user that a timelapse is still ongoing.
            #
            # TODO remove when central job management becomes available and takes care of this for us
            self.sendEvent(
                Events.MOVIE_RENDERING, payload=octoprint.timelapse.current_render_job
            )
        self._registered = True

    def _unregister(self):
        """Unregister this socket from the system"""

        self._printer.unregister_callback(self)
        self._fileManager.unregister_slicingprogress_callback(self)
        octoprint.timelapse.unregister_callback(self)
        for event in octoprint.events.all_events():
            self._eventManager.unsubscribe(event, self._onEvent)

    def _reregister(self):
        """Unregister and register again"""
        self._unregister()
        self._register()

    def _sendReauthRequired(self, reason):
        self._emit("reauthRequired", payload={"reason": reason})

    def _emit(self, type, payload=None, permissions=None):
        proceed = True
        for name, hook in self._emit_hooks.items():
            try:
                proceed = proceed and hook(self, self._user, type, payload)
            except Exception:
                self._logger.exception(
                    f"Error processing emit hook handler from plugin {name}",
                    extra={"plugin": name},
                )

        if not proceed:
            return

        if permissions is None:
            permissions = self._emit_permissions.get(type, self._emit_permissions["*"])
            permissions = (
                permissions(payload)
                if callable(permissions)
                else [x for x in permissions]
            )

        if not self._user or not all(
            map(lambda p: self._user.has_permission(p), permissions)
        ):
            if not self._authed:
                with self._unauthed_backlog_mutex:
                    if len(self._unauthed_backlog) < self._unauthed_backlog_max:
                        self._unauthed_backlog.append((type, payload))
                        self._logger.debug(
                            "Socket message held back until permissions cleared, added to backlog: {}".format(
                                type
                            )
                        )
                    else:
                        self._logger.debug(
                            "Socket message held back, but backlog full. Throwing message away: {}".format(
                                type
                            )
                        )
            return

        self._do_emit(type, payload)

    def _do_emit(self, type, payload):
        try:
            self.send({type: payload})
        except Exception as e:
            if self._logger.isEnabledFor(logging.DEBUG):
                self._logger.exception(
                    f"Could not send message to client {self._remoteAddress}"
                )
            else:
                self._logger.warning(
                    "Could not send message to client {}: {}".format(
                        self._remoteAddress, e
                    )
                )

    def _on_login(self, user):
        self._user = user
        self._logger.info(
            "User {} logged in on the socket from client {}".format(
                user.get_name(), self._remoteAddress
            )
        )
        self._authed = True

        self._keep_alive.start()

        for name, hook in self._authed_hooks.items():
            try:
                hook(self, self._user)
            except Exception:
                self._logger.exception(
                    f"Error processing authed hook handler for plugin {name}",
                    extra={"plugin": name},
                )

        # if we have a backlog from being unauthed, process that now
        with self._unauthed_backlog_mutex:
            backlog = self._unauthed_backlog
            self._unauthed_backlog = []

        if len(backlog):
            self._logger.debug(
                "Sending {} messages on the socket that were held back".format(
                    len(backlog)
                )
            )
            for message, payload in backlog:
                self._do_emit(message, payload)

        # trigger ClientAuthed event
        octoprint.events.eventManager().fire(
            octoprint.events.Events.CLIENT_AUTHED,
            payload={"username": user.get_name(), "remoteAddress": self._remoteAddress},
        )

    def _on_logout(self):
        self._user = self._userManager.anonymous_user_factory()
        self._authed = False

        for name, hook in self._authed_hooks.items():
            try:
                hook(self, self._user)
            except Exception:
                self._logger.exception(
                    f"Error processing authed hook handler for plugin {name}",
                    extra={"plugin": name},
                )



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

import threading

import requests
from flask_babel import gettext

import octoprint.plugin
from octoprint.schema.webcam import RatioEnum, Webcam, WebcamCompatibility
from octoprint.webcams import WebcamNotAbleToTakeSnapshotException


class ClassicWebcamPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.WebcamProviderPlugin,
    octoprint.plugin.WizardPlugin,
):
    def __init__(self):
        self._capture_mutex = threading.Lock()
        self._webcam_name = "classic"

    # ~~ TemplatePlugin API

    def get_assets(self):
        return {
            "js": [
                "js/classicwebcam.js",
                "js/classicwebcam_settings.js",
                "js/classicwebcam_wizard.js",
            ],
            "less": ["less/classicwebcam.less"],
            "css": ["css/classicwebcam.css"],
        }

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "template": "classicwebcam_settings.jinja2",
                "custom_bindings": True,
            },
            {
                "type": "webcam",
                "name": "Classic Webcam",
                "template": "classicwebcam_webcam.jinja2",
                "custom_bindings": True,
                "suffix": "_real",
            },
            {
                "type": "wizard",
                "name": "Classic Webcam Wizard",
                "template": "classicwebcam_wizard.jinja2",
                "suffix": "_wizard",
            },
        ]

    # ~~ WebcamProviderPlugin API

    def get_webcam_configurations(self):
        streamRatio = self._settings.get(["streamRatio"])
        if streamRatio == "4:3":
            streamRatio = RatioEnum.four_three
        else:
            streamRatio = RatioEnum.sixteen_nine
        webRtcServers = self._settings.get(["streamWebrtcIceServers"])
        cacheBuster = self._settings.get_boolean(["cacheBuster"])
        stream = self._get_stream_url()
        snapshot = self._get_snapshot_url()
        flipH = self._settings.get_boolean(["flipH"])
        flipV = self._settings.get_boolean(["flipV"])
        rotate90 = self._settings.get_boolean(["rotate90"])
        snapshotSslValidation = self._settings.get_boolean(["snapshotSslValidation"])

        try:
            streamTimeout = int(self._settings.get(["streamTimeout"]))
        except Exception:
            streamTimeout = 5

        try:
            snapshotTimeout = int(self._settings.get(["snapshotTimeout"]))
        except Exception:
            snapshotTimeout = 5

        return [
            Webcam(
                name=self._webcam_name,
                displayName="Classic Webcam",
                flipH=flipH,
                flipV=flipV,
                rotate90=rotate90,
                snapshotDisplay=snapshot,
                canSnapshot=self._can_snapshot(),
                compat=WebcamCompatibility(
                    stream=stream,
                    streamTimeout=streamTimeout,
                    streamRatio=streamRatio,
                    cacheBuster=cacheBuster,
                    streamWebrtcIceServers=webRtcServers,
                    snapshot=snapshot,
                    snapshotTimeout=snapshotTimeout,
                    snapshotSslValidation=snapshotSslValidation,
                ),
                extras=dict(
                    stream=stream,
                    streamTimeout=streamTimeout,
                    streamRatio=streamRatio,
                    streamWebrtcIceServers=webRtcServers,
                    cacheBuster=cacheBuster,
                ),
            ),
        ]

    def _get_snapshot_url(self):
        return self._settings.get(["snapshot"])

    def _get_stream_url(self):
        return self._settings.get(["stream"])

    def _can_snapshot(self):
        snapshot = self._get_snapshot_url()
        return snapshot is not None and snapshot.strip() != ""

    def take_webcam_snapshot(self, _):
        snapshot_url = self._get_snapshot_url()
        if not self._can_snapshot():
            raise WebcamNotAbleToTakeSnapshotException(self._webcam_name)

        with self._capture_mutex:
            self._logger.debug(f"Capturing image from {snapshot_url}")
            r = requests.get(
                snapshot_url,
                stream=True,
                timeout=self._settings.get_int(["snapshotTimeout"]),
                verify=self._settings.get_boolean(["snapshotSslValidation"]),
            )
            r.raise_for_status()
            return r.iter_content(chunk_size=1024)

    # ~~ SettingsPlugin API

    def get_settings_defaults(self):
        return dict(
            flipH=False,
            flipV=False,
            rotate90=False,
            stream="",
            streamTimeout=5,
            streamRatio="16:9",
            streamWebrtcIceServers=["stun:stun.l.google.com:19302"],
            snapshot="",
            cacheBuster=False,
            snapshotSslValidation=True,
            snapshotTimeout=5,
        )

    def get_settings_version(self):
        return 1

    def on_settings_migrate(self, target, current):
        if current is None:
            config = self._settings.global_get(["webcam"])
            if config:
                self._logger.info(
                    "Migrating settings from webcam to plugins.classicwebcam..."
                )

                # flipH
                self._settings.set_boolean(["flipH"], config.get("flipH", False))
                self._settings.global_remove(["webcam", "flipH"])

                # flipV
                self._settings.set_boolean(["flipV"], config.get("flipV", False))
                self._settings.global_remove(["webcam", "flipV"])

                # rotate90
                self._settings.set_boolean(["rotate90"], config.get("rotate90", False))
                self._settings.global_remove(["webcam", "rotate90"])

                # stream
                self._settings.set(["stream"], config.get("stream", ""))
                self._settings.global_remove(["webcam", "stream"])

                # streamTimeout
                self._settings.set_int(["streamTimeout"], config.get("streamTimeout", 5))
                self._settings.global_remove(["webcam", "streamTimeout"])

                # streamRatio
                self._settings.set(["streamRatio"], config.get("streamRatio", "16:9"))
                self._settings.global_remove(["webcam", "streamRatio"])

                # streamWebrtcIceServers
                self._settings.set(
                    ["streamWebrtcIceServers"],
                    config.get(
                        "streamWebrtcIceServers", ["stun:stun.l.google.com:19302"]
                    ),
                )
                self._settings.global_remove(["webcam", "streamWebrtcIceServers"])

                # snapshot
                self._settings.set(["snapshot"], config.get("snapshot", ""))
                self._settings.global_remove(["webcam", "snapshot"])

                # cacheBuster
                self._settings.set_boolean(
                    ["cacheBuster"], config.get("cacheBuster", False)
                )
                self._settings.global_remove(["webcam", "cacheBuster"])

                # snapshotTimeout
                self._settings.set_int(
                    ["snapshotTimeout"], config.get("snapshotTimeout", 5)
                )
                self._settings.global_remove(["webcam", "snapshotTimeout"])

                # snapshotSslValidation
                self._settings.set_boolean(
                    ["snapshotSslValidation"], config.get("snapshotSslValidation", True)
                )
                self._settings.global_remove(["webcam", "snapshotSslValidation"])

    # ~~ WizardPlugin API

    def is_wizard_required(self):
        required = (
            not self._get_stream_url()
            or not self._get_snapshot_url()
            or not self._settings.global_get(["webcam", "ffmpegPath"])
        )
        firstrun = self._settings.global_get(["server", "firstRun"])
        return required and firstrun

    def get_wizard_version(self):
        return 1


__plugin_name__ = gettext("Classic Webcam")
__plugin_author__ = "Christian Würthner"
__plugin_description__ = "Provides a simple webcam viewer in OctoPrint's UI, images provided by an MJPEG webcam."
__plugin_disabling_discouraged__ = gettext(
    "This plugin provides the standard webcam in OctoPrint. If you do not have any other plugin providing a webcam set up, the webcam section in the control tab will no longer be visible."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = ClassicWebcamPlugin()



from flask_babel import gettext

import octoprint.events
import octoprint.plugin
from octoprint.access import ADMIN_GROUP


class EventManagerPlugin(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
):
    def on_settings_load(self):
        my_settings = {
            "availableEvents": octoprint.events.all_events(),
            "subscriptions": [],
        }
        events = self._settings.global_get(["events"])
        if events:
            my_settings["subscriptions"] = sorted(
                events.get("subscriptions", []), key=(lambda x: x["event"])
            )
        return my_settings

    def on_settings_save(self, data):
        if type(data.get("subscriptions")) == list:
            self._settings.global_set(
                ["events", "subscriptions"], data.get("subscriptions", [])
            )

    def get_assets(self):
        return {"js": ["js/events.js"]}

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "custom_bindings": True,
                "name": gettext("Event Manager"),
            }
        ]

    def get_additional_permissions(self):
        return [
            {
                "key": "MANAGE",
                "name": "Event management",
                "description": gettext(
                    "Allows for the management of event subscriptions."
                ),
                "default_groups": [ADMIN_GROUP],
                "roles": ["manage"],
            }
        ]


__plugin_name__ = gettext("Event Manager")
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_author__ = "jneilliii"
__plugin_license__ = "AGPLv3"
__plugin_description__ = gettext(
    "Plugin to configure event subscriptions available in config.yaml."
)
__plugin_implementation__ = EventManagerPlugin()



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin
from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions
from octoprint.events import Events
from octoprint.server import NO_CONTENT
from octoprint.server.util.flask import no_firstrun_access
from octoprint.settings import default_settings
from octoprint.util import is_hidden_path, to_bytes, yaml
from octoprint.util.pip import create_pip_caller
from octoprint.util.platform import is_os_compatible
from octoprint.util.version import (
    get_comparable_version,
    get_octoprint_version,
    get_octoprint_version_string,
    is_octoprint_compatible,
)

try:
    import zlib  # check if zlib is available
except ImportError:
    zlib = None


import json
import logging
import os
import shutil
import sys
import tempfile
import threading
import time
import traceback
import zipfile

import flask
import requests
import sarge
from flask_babel import gettext

from octoprint.plugins.pluginmanager import DEFAULT_PLUGIN_REPOSITORY
from octoprint.settings import valid_boolean_trues
from octoprint.util import get_formatted_size
from octoprint.util.text import sanitize

UNKNOWN_PLUGINS_FILE = "unknown_plugins_from_restore.json"

BACKUP_DATE_TIME_FMT = "%Y%m%d-%H%M%S"

MAX_UPLOAD_SIZE = 1024 * 1024 * 1024  # 1GB


class BackupPlugin(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.WizardPlugin,
):
    _pip_caller = None

    _backup_in_progress = threading.Lock()
    _restore_in_progress = threading.Lock()

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "ACCESS",
                "name": "Backup access",
                "description": gettext("Allows access to backups and restores"),
                "roles": ["access"],
                "dangerous": True,
                "default_groups": [ADMIN_GROUP],
            }
        ]

    # Socket emit hook

    def socket_emit_hook(self, socket, user, message, payload, *args, **kwargs):
        if message != "event" or payload["type"] != Events.PLUGIN_BACKUP_BACKUP_CREATED:
            return True

        return user and user.has_permission(Permissions.PLUGIN_BACKUP_ACCESS)

    ##~~ StartupPlugin

    def on_after_startup(self):
        self._clean_dir_backup(self._settings._basedir, on_log_progress=self._logger.info)

    ##~~ SettingsPlugin

    def get_settings_defaults(self):
        return {"restore_unsupported": False}

    ##~~ AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/backup.js"],
            "clientjs": ["clientjs/backup.js"],
            "css": ["css/backup.css"],
            "less": ["less/backup.less"],
        }

    ##~~ TemplatePlugin

    def get_template_configs(self):
        return [
            {"type": "settings", "name": gettext("Backup & Restore")},
            {"type": "wizard", "name": gettext("Restore Backup?")},
        ]

    def get_template_vars(self):
        return {
            "max_upload_size": MAX_UPLOAD_SIZE,
            "max_upload_size_str": get_formatted_size(MAX_UPLOAD_SIZE),
        }

    ##~~ BlueprintPlugin

    @octoprint.plugin.BlueprintPlugin.route("/", methods=["GET"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def get_state(self):
        backups = self._get_backups()
        unknown_plugins = self._get_unknown_plugins()
        return flask.jsonify(
            backups=backups,
            backup_in_progress=self._backup_in_progress.locked(),
            unknown_plugins=unknown_plugins,
            restore_supported=self._restore_supported(self._settings),
            max_upload_size=MAX_UPLOAD_SIZE,
        )

    @octoprint.plugin.BlueprintPlugin.route("/unknown_plugins", methods=["GET"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def get_unknown_plugins(self):
        # TODO add caching
        unknown_plugins = self._get_unknown_plugins()
        return flask.jsonify(unknown_plugins=unknown_plugins)

    @octoprint.plugin.BlueprintPlugin.route("/unknown_plugins", methods=["DELETE"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def delete_unknown_plugins(self):
        data_file = os.path.join(self.get_plugin_data_folder(), UNKNOWN_PLUGINS_FILE)
        try:
            os.remove(data_file)
        except Exception:
            pass
        return NO_CONTENT

    @octoprint.plugin.BlueprintPlugin.route("/backup", methods=["GET"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def get_backups(self):
        backups = self._get_backups()
        return flask.jsonify(backups=backups)

    @octoprint.plugin.BlueprintPlugin.route("/backup", methods=["POST"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def create_backup(self):
        data = flask.request.json
        exclude = data.get("exclude", [])
        filename = self._build_backup_filename(settings=self._settings)

        self._start_backup(exclude, filename)

        response = flask.jsonify(started=True, name=filename)
        response.status_code = 201
        return response

    @octoprint.plugin.BlueprintPlugin.route("/backup/<filename>", methods=["DELETE"])
    @no_firstrun_access
    @Permissions.PLUGIN_BACKUP_ACCESS.require(403)
    def delete_backup(self, filename):
        self._delete_backup(filename)

        return NO_CONTENT

    @octoprint.plugin.BlueprintPlugin.route("/restore", methods=["POST"])
    def perform_restore(self):
        if not Permissions.PLUGIN_BACKUP_ACCESS.can() and not self._settings.global_get(
            ["server", "firstRun"]
        ):
            flask.abort(403)

        if not self._restore_supported(self._settings):
            flask.abort(
                400,
                description="Invalid request, the restores are not "
                "supported on the underlying operating system",
            )

        input_name = "file"
        input_upload_path = (
            input_name
            + "."
            + self._settings.global_get(["server", "uploads", "pathSuffix"])
        )

        if input_upload_path in flask.request.values:
            # file to restore was uploaded
            path = flask.request.values[input_upload_path]

        elif flask.request.json and "path" in flask.request.json:
            # existing backup is supposed to be restored
            backup_folder = self.get_plugin_data_folder()
            path = os.path.realpath(
                os.path.join(backup_folder, flask.request.json["path"])
            )
            if (
                not path.startswith(backup_folder)
                or not os.path.exists(path)
                or is_hidden_path(path)
            ):
                return flask.abort(404)

        else:
            flask.abort(
                400,
                description="Invalid request, neither a file nor a path of a file to "
                "restore provided",
            )

        def on_install_plugins(plugins):
            force_user = self._settings.global_get_boolean(
                ["plugins", "pluginmanager", "pip_force_user"]
            )
            pip_args = self._settings.global_get(["plugins", "pluginmanager", "pip_args"])

            def on_log(line):
                self._logger.info(line)
                self._send_client_message("logline", {"line": line, "type": "stdout"})

            for plugin in plugins:
                octoprint_compatible = is_octoprint_compatible(
                    *plugin["compatibility"]["octoprint"]
                )
                os_compatible = is_os_compatible(plugin["compatibility"]["os"])
                compatible = octoprint_compatible and os_compatible
                if not compatible:
                    if not octoprint_compatible and not os_compatible:
                        self._logger.warning(
                            "Cannot install plugin {}, it is incompatible to this version "
                            "of OctoPrint and the underlying operating system".format(
                                plugin["id"]
                            )
                        )
                    elif not octoprint_compatible:
                        self._logger.warning(
                            "Cannot install plugin {}, it is incompatible to this version "
                            "of OctoPrint".format(plugin["id"])
                        )
                    elif not os_compatible:
                        self._logger.warning(
                            "Cannot install plugin {}, it is incompatible to the underlying "
                            "operating system".format(plugin["id"])
                        )
                    self._send_client_message(
                        "plugin_incompatible",
                        {
                            "plugin": plugin["id"],
                            "octoprint_compatible": octoprint_compatible,
                            "os_compatible": os_compatible,
                        },
                    )
                    continue

                self._logger.info("Installing plugin {}".format(plugin["id"]))
                self._send_client_message("installing_plugin", {"plugin": plugin["id"]})
                self.__class__._install_plugin(
                    plugin,
                    force_user=force_user,
                    pip_command=self._settings.global_get(
                        ["server", "commands", "localPipCommand"]
                    ),
                    pip_args=pip_args,
                    on_log=on_log,
                )

        def on_report_unknown_plugins(plugins):
            self._send_client_message("unknown_plugins", payload={"plugins": plugins})

        def on_log_progress(line):
            self._logger.info(line)
            self._send_client_message(
                "logline", payload={"line": line, "stream": "stdout"}
            )

        def on_log_error(line, exc_info=None):
            self._logger.error(line, exc_info=exc_info)
            self._send_client_message(
                "logline", payload={"line": line, "stream": "stderr"}
            )

            if exc_info is not None:
                exc_type, exc_value, exc_tb = exc_info
                output = traceback.format_exception(exc_type, exc_value, exc_tb)
                for line in output:
                    self._send_client_message(
                        "logline", payload={"line": line.rstrip(), "stream": "stderr"}
                    )

        def on_restore_start(path):
            self._send_client_message("restore_started")

        def on_restore_done(path):
            self._send_client_message("restore_done")

        def on_restore_failed(path):
            self._send_client_message("restore_failed")

        def on_invalid_backup(line):
            on_log_error(line)

        archive = tempfile.NamedTemporaryFile(delete=False)
        archive.close()
        shutil.copy(path, archive.name)
        path = archive.name

        # noinspection PyTypeChecker
        thread = threading.Thread(
            target=self._restore_backup,
            args=(path,),
            kwargs={
                "settings": self._settings,
                "plugin_manager": self._plugin_manager,
                "datafolder": self.get_plugin_data_folder(),
                "on_install_plugins": on_install_plugins,
                "on_report_unknown_plugins": on_report_unknown_plugins,
                "on_invalid_backup": on_invalid_backup,
                "on_log_progress": on_log_progress,
                "on_log_error": on_log_error,
                "on_restore_start": on_restore_start,
                "on_restore_done": on_restore_done,
                "on_restore_failed": on_restore_failed,
            },
        )
        thread.daemon = True
        thread.start()

        return flask.jsonify(started=True)

    def is_blueprint_protected(self):
        return False

    def is_blueprint_csrf_protected(self):
        return True

    ##~~ WizardPlugin

    def is_wizard_required(self):
        return self._settings.global_get(["server", "firstRun"]) and is_os_compatible(
            ["!windows"]
        )

    def get_wizard_details(self):
        return {"required": self.is_wizard_required()}

    ##~~ tornado hook

    def route_hook(self, *args, **kwargs):
        from octoprint.server import app
        from octoprint.server.util.flask import permission_validator
        from octoprint.server.util.tornado import (
            LargeResponseHandler,
            access_validation_factory,
            path_validation_factory,
        )
        from octoprint.util import is_hidden_path

        plugin_folder = self.get_plugin_data_folder()

        def path_check(path):
            joined = os.path.join(plugin_folder, path)
            return not is_hidden_path(joined) and self._valid_backup(joined)

        return [
            (
                r"/download/(.*)",
                LargeResponseHandler,
                {
                    "path": plugin_folder,
                    "as_attachment": True,
                    "path_validation": path_validation_factory(
                        path_check,
                        status_code=404,
                    ),
                    "access_validation": access_validation_factory(
                        app, permission_validator, Permissions.PLUGIN_BACKUP_ACCESS
                    ),
                },
            )
        ]

    def bodysize_hook(self, current_max_body_sizes, *args, **kwargs):
        # max upload size for the restore endpoint
        return [("POST", r"/restore", MAX_UPLOAD_SIZE)]

    # Exported plugin helpers
    def create_backup_helper(self, exclude=None, filename=None):
        """
        .. versionadded:: 1.6.0

        Create a backup from a plugin or other internal call

        This helper is exported as ``create_backup`` and can be used from the plugin
        manager's ``get_helpers`` method.

        **Example**

        The following code snippet can be used from within a plugin, and will create a backup
        excluding two folders (``timelapse`` and ``uploads``)

        .. code-block:: python

            helpers = self._plugin_manager.get_helpers("backup", "create_backup")

            if helpers and "create_backup" in helpers:
                helpers["create_backup"](exclude=["timelapse", "uploads"])

        By using the ``if helpers [...]`` clause, plugins can fall back to other methods
        when they are running under versions where these helpers did not exist.


        :param list exclude: Names of data folders to exclude, defaults to None
        :param str filename: Name of backup to be created, if None (default) the backup
            name will be auto-generated. This should use a ``.zip`` extension.
        """
        if exclude is None:
            exclude = []
        if not isinstance(exclude, list):
            exclude = list(exclude)

        self._start_backup(exclude, filename=filename)

    def delete_backup_helper(self, filename):
        """
        .. versionadded:: 1.6.0

        Delete the specified backup from a plugin or other internal call

        This helper is exported as ``delete_backup`` and can be used through the plugin
        manager's ``get_helpers`` method.

        **Example**
        The following code snippet can be used from within a plugin, and will attempt to
        delete the backup named ``ExampleBackup.zip``.

        .. code-block:: python

            helpers = self._plugin_manager.get_helpers("backup", "delete_backup")

            if helpers and "delete_backup" in helpers:
                helpers["delete_backup"]("ExampleBackup.zip")

        By using the ``if helpers [...]`` clause, plugins can fall back to other methods
        when they are running under versions where these helpers did not exist.

        .. warning::

            This method will fail silently if the backup does not exist, and so
            it is recommended that you make sure the name comes from a verified source,
            for example the name from the events or other helpers.

        :param str filename: The name of the backup to delete
        """
        self._delete_backup(filename)

    ##~~ CLI hook

    def cli_commands_hook(self, cli_group, pass_octoprint_ctx, *args, **kwargs):
        import click

        @click.command("backup")
        @click.option(
            "--exclude",
            multiple=True,
            help="Identifiers of data folders to exclude, e.g. 'uploads' to exclude uploads or "
            "'timelapse' to exclude timelapses.",
        )
        @click.option(
            "--path",
            type=click.Path(),
            default=None,
            help="Specify full path to backup file to be created",
        )
        def backup_command(exclude, path):
            """
            Creates a new backup.
            """

            settings = octoprint.plugin.plugin_settings_for_settings_plugin(
                "backup", self, settings=cli_group.settings
            )

            if path is not None:
                datafolder, filename = os.path.split(os.path.abspath(path))
            else:
                filename = None
                datafolder = os.path.join(settings.getBaseFolder("data"), "backup")

            if not os.path.isdir(datafolder):
                os.makedirs(datafolder)

            def on_backup_start(name, temporary_path, exclude):
                click.echo(f"Creating backup at {name}, please wait...")

            path = self._create_backup(
                name=filename,
                exclude=exclude,
                settings=settings,
                plugin_manager=cli_group.plugin_manager,
                datafolder=datafolder,
                on_backup_start=on_backup_start,
            )
            click.echo("Done.")
            click.echo(f"Backup located at {path}")

        @click.command("restore")
        @click.argument("path")
        def restore_command(path):
            """
            Restores an existing backup from the backup zip provided as argument.

            OctoPrint does not need to run for this to proceed.
            """
            settings = octoprint.plugin.plugin_settings_for_settings_plugin(
                "backup", self, settings=cli_group.settings
            )
            plugin_manager = cli_group.plugin_manager

            datafolder = os.path.join(settings.getBaseFolder("data"), "backup")
            if not os.path.isdir(datafolder):
                os.makedirs(datafolder)

            # register plugin manager plugin setting overlays
            plugin_info = plugin_manager.get_plugin_info("pluginmanager")
            if plugin_info and plugin_info.implementation:
                default_settings_overlay = {"plugins": {}}
                default_settings_overlay["plugins"][
                    "pluginmanager"
                ] = plugin_info.implementation.get_settings_defaults()
                settings.add_overlay(default_settings_overlay, at_end=True)

            if not os.path.isabs(path):
                datafolder = os.path.join(settings.getBaseFolder("data"), "backup")
                if not os.path.isdir(datafolder):
                    os.makedirs(datafolder)
                path = os.path.join(datafolder, path)

            if not os.path.exists(path):
                click.echo(f"Backup {path} does not exist", err=True)
                sys.exit(-1)

            archive = tempfile.NamedTemporaryFile(delete=False)
            archive.close()
            shutil.copy(path, archive.name)
            path = archive.name

            def on_install_plugins(plugins):
                if not plugins:
                    return

                force_user = settings.global_get_boolean(
                    ["plugins", "pluginmanager", "pip_force_user"]
                )
                pip_args = settings.global_get(["plugins", "pluginmanager", "pip_args"])

                def log(line):
                    click.echo(f"\t{line}")

                for plugin in plugins:
                    octoprint_compatible = is_octoprint_compatible(
                        *plugin["compatibility"]["octoprint"]
                    )
                    os_compatible = is_os_compatible(plugin["compatibility"]["os"])
                    compatible = octoprint_compatible and os_compatible
                    if not compatible:
                        if not octoprint_compatible and not os_compatible:
                            click.echo(
                                "Cannot install plugin {}, it is incompatible to this version of "
                                "OctoPrint and the underlying operating system".format(
                                    plugin["id"]
                                )
                            )
                        elif not octoprint_compatible:
                            click.echo(
                                "Cannot install plugin {}, it is incompatible to this version of "
                                "OctoPrint".format(plugin["id"])
                            )
                        elif not os_compatible:
                            click.echo(
                                "Cannot install plugin {}, it is incompatible to the underlying "
                                "operating system".format(plugin["id"])
                            )
                        continue

                    click.echo("Installing plugin {}".format(plugin["id"]))
                    self.__class__._install_plugin(
                        plugin,
                        force_user=force_user,
                        pip_command=settings.global_get(
                            ["server", "commands", "localPipCommand"]
                        ),
                        pip_args=pip_args,
                        on_log=log,
                    )

            def on_report_unknown_plugins(plugins):
                if not plugins:
                    return

                click.echo(
                    "The following plugins were not found in the plugin repository. You'll need to install them manually."
                )
                for plugin in plugins:
                    click.echo(
                        "\t{} (Homepage: {})".format(
                            plugin["name"], plugin["url"] if plugin["url"] else "?"
                        )
                    )

            def on_log_progress(line):
                click.echo(line)

            def on_log_error(line, exc_info=None):
                click.echo(line, err=True)

                if exc_info is not None:
                    exc_type, exc_value, exc_tb = exc_info
                    output = traceback.format_exception(exc_type, exc_value, exc_tb)
                    for line in output:
                        click.echo(line.rstrip(), err=True)

            if self._restore_backup(
                path,
                settings=settings,
                plugin_manager=plugin_manager,
                datafolder=datafolder,
                on_install_plugins=on_install_plugins,
                on_report_unknown_plugins=on_report_unknown_plugins,
                on_log_progress=on_log_progress,
                on_log_error=on_log_error,
                on_invalid_backup=on_log_error,
            ):
                click.echo(f"Restored from {path}")
            else:
                click.echo(f"Restoring from {path} failed", err=True)

        return [backup_command, restore_command]

    ##~~ helpers

    def _start_backup(self, exclude, filename=None):
        def on_backup_start(name, temporary_path, exclude):
            self._logger.info(
                "Creating backup zip at {} (excluded: {})...".format(
                    temporary_path, ",".join(exclude) if len(exclude) else "-"
                )
            )
            self._send_client_message("backup_started", payload={"name": name})

        def on_backup_done(name, final_path, exclude):
            self._send_client_message("backup_done", payload={"name": name})
            self._logger.info("... done creating backup zip.")

            self._event_bus.fire(
                Events.PLUGIN_BACKUP_BACKUP_CREATED,
                {"name": name, "path": final_path, "excludes": exclude},
            )

        def on_backup_error(name, exc_info):
            self._send_client_message(
                "backup_error", payload={"name": name, "error": f"{exc_info[1]}"}
            )
            self._logger.error("Error while creating backup zip", exc_info=exc_info)

        thread = threading.Thread(
            target=self._create_backup,
            kwargs={
                "name": filename,
                "exclude": exclude,
                "settings": self._settings,
                "plugin_manager": self._plugin_manager,
                "logger": self._logger,
                "datafolder": self.get_plugin_data_folder(),
                "on_backup_start": on_backup_start,
                "on_backup_done": on_backup_done,
                "on_backup_error": on_backup_error,
            },
        )
        thread.daemon = True
        thread.start()

    def _delete_backup(self, filename):
        """
        Delete the backup specified
        Args:
            filename (str): Name of backup to delete
        """
        backup_folder = self.get_plugin_data_folder()
        full_path = os.path.realpath(os.path.join(backup_folder, filename))
        if (
            full_path.startswith(backup_folder)
            and os.path.exists(full_path)
            and not is_hidden_path(full_path)
        ):
            try:
                os.remove(full_path)
            except Exception:
                self._logger.exception(f"Could not delete {filename}")
                raise

    def _get_backups(self):
        backups = []
        for entry in os.scandir(self.get_plugin_data_folder()):
            if is_hidden_path(entry.path):
                continue
            if not self._valid_backup(entry.path):
                continue

            backups.append(
                {
                    "name": entry.name,
                    "date": entry.stat().st_mtime,
                    "size": entry.stat().st_size,
                    "url": flask.url_for("index")
                    + "plugin/backup/download/"
                    + entry.name,
                }
            )
        return backups

    def _get_unknown_plugins(self):
        data_file = os.path.join(self.get_plugin_data_folder(), UNKNOWN_PLUGINS_FILE)
        if os.path.exists(data_file):
            try:
                with open(data_file, encoding="utf-8") as f:
                    unknown_plugins = json.load(f)

                assert isinstance(unknown_plugins, list)
                assert all(
                    map(
                        lambda x: isinstance(x, dict)
                        and "key" in x
                        and "name" in x
                        and "url" in x,
                        unknown_plugins,
                    )
                )

                installed_plugins = self._plugin_manager.plugins
                unknown_plugins = list(
                    filter(lambda x: x["key"] not in installed_plugins, unknown_plugins)
                )
                if not unknown_plugins:
                    # no plugins left uninstalled, delete data file
                    try:
                        os.remove(data_file)
                    except Exception:
                        self._logger.exception(
                            "Error while deleting list of unknown plugins at {}".format(
                                data_file
                            )
                        )

                return unknown_plugins
            except Exception:
                self._logger.exception(
                    "Error while reading list of unknown plugins from {}".format(
                        data_file
                    )
                )
                try:
                    os.remove(data_file)
                except Exception:
                    self._logger.exception(
                        "Error while deleting list of unknown plugins at {}".format(
                            data_file
                        )
                    )

        return []

    @classmethod
    def _clean_dir_backup(cls, basedir, on_log_progress=None):
        basedir_backup = basedir + ".bck"

        if os.path.exists(basedir_backup):

            def remove_bck():
                if callable(on_log_progress):
                    on_log_progress(
                        "Found config folder backup from prior restore, deleting it..."
                    )
                shutil.rmtree(basedir_backup)
                if callable(on_log_progress):
                    on_log_progress("... deleted.")

            thread = threading.Thread(target=remove_bck)
            thread.daemon = True
            thread.start()

    @classmethod
    def _get_disk_size(cls, path, ignored=None):
        if ignored is None:
            ignored = []

        if path in ignored:
            return 0

        total = 0
        for entry in os.scandir(path):
            if entry.is_dir():
                total += cls._get_disk_size(entry.path, ignored=ignored)
            elif entry.is_file():
                total += entry.stat().st_size
        return total

    @classmethod
    def _free_space(cls, path, size):
        from psutil import disk_usage

        return disk_usage(path).free > size

    @classmethod
    def _get_plugin_repository_data(cls, url, logger=None):
        if logger is None:
            logger = logging.getLogger(__name__)

        try:
            r = requests.get(url, timeout=3.05)
            r.raise_for_status()
        except Exception:
            logger.exception(
                f"Error while fetching the plugin repository data from {url}"
            )
            return {}

        from octoprint.plugins.pluginmanager import map_repository_entry

        return {plugin["id"]: plugin for plugin in map(map_repository_entry, r.json())}

    @classmethod
    def _install_plugin(
        cls, plugin, force_user=False, pip_command=None, pip_args=None, on_log=None
    ):
        if pip_args is None:
            pip_args = []

        if on_log is None:
            on_log = logging.getLogger(__name__).info

        # prepare pip caller
        def log(prefix, *lines):
            for line in lines:
                on_log(f"{prefix} {line.rstrip()}")

        def log_call(*lines):
            log(">", *lines)

        def log_stdout(*lines):
            log("<", *lines)

        def log_stderr(*lines):
            log("!", *lines)

        if cls._pip_caller is None:
            cls._pip_caller = create_pip_caller(
                command=pip_command, force_user=force_user
            )

        cls._pip_caller.on_log_call = log_call
        cls._pip_caller.on_log_stdout = log_stdout
        cls._pip_caller.on_log_stderr = log_stderr

        # install plugin
        pip = ["install", sarge.shell_quote(plugin["archive"]), "--no-cache-dir"]

        if plugin.get("follow_dependency_links"):
            pip.append("--process-dependency-links")

        if force_user:
            pip.append("--user")

        if pip_args:
            pip += pip_args

        cls._pip_caller.execute(*pip)

    @classmethod
    def _create_backup(
        cls,
        name=None,
        exclude=None,
        settings=None,
        plugin_manager=None,
        logger=None,
        datafolder=None,
        on_backup_start=None,
        on_backup_done=None,
        on_backup_error=None,
    ):
        if logger is None:
            logger = logging.getLogger(__name__)

        exclude_by_default = (
            "generated",
            "logs",
            "watched",
        )

        with cls._backup_in_progress:
            if name is None:
                name = cls._build_backup_filename(settings)

            try:
                if exclude is None:
                    exclude = []
                if not isinstance(exclude, list):
                    exclude = list(exclude)

                if "timelapse" in exclude:
                    exclude.append("timelapse_tmp")

                current_excludes = list(exclude)
                additional_excludes = list()
                plugin_data = settings.global_get_basefolder("data")
                for plugin, hook in plugin_manager.get_hooks(
                    "octoprint.plugin.backup.additional_excludes"
                ).items():
                    try:
                        additional = hook(current_excludes)
                        if isinstance(additional, list):
                            if "." in additional:
                                current_excludes.append(os.path.join("data", plugin))
                                additional_excludes.append(
                                    os.path.join(plugin_data, plugin)
                                )
                            else:
                                current_excludes += map(
                                    lambda x: os.path.join("data", plugin, x), additional
                                )
                                additional_excludes += map(
                                    lambda x: os.path.join(plugin_data, plugin, x),
                                    additional,
                                )
                    except Exception:
                        logger.exception(
                            f"Error while retrieving additional excludes "
                            f"from plugin {plugin}",
                            extra={"plugin": plugin},
                        )

                configfile = settings._configfile
                basedir = settings._basedir

                temporary_path = os.path.join(datafolder, f".{name}")
                final_path = os.path.join(datafolder, name)

                own_folder = datafolder
                defaults = [
                    os.path.join(basedir, "config.yaml"),
                ] + [
                    os.path.join(basedir, folder)
                    for folder in default_settings["folder"].keys()
                ]

                # check how many bytes we are about to backup
                size = os.stat(configfile).st_size
                for folder in default_settings["folder"].keys():
                    if folder in exclude or folder in exclude_by_default:
                        continue
                    size += cls._get_disk_size(
                        settings.global_get_basefolder(folder),
                        ignored=[
                            own_folder,
                        ],
                    )
                size += cls._get_disk_size(
                    basedir,
                    ignored=defaults
                    + [
                        own_folder,
                    ],
                )

                # since we can't know the compression ratio beforehand, we assume we need the same amount of space
                if not cls._free_space(os.path.dirname(temporary_path), size):
                    raise InsufficientSpace()

                compression = zipfile.ZIP_DEFLATED if zlib else zipfile.ZIP_STORED

                backup_error = False
                try:
                    for plugin, hook in plugin_manager.get_hooks(
                        "octoprint.plugin.backup.before_backup"
                    ).items():
                        try:
                            hook()
                        except Exception:
                            logger.exception(
                                f"Error while running before_backup hook from plugin {plugin}",
                                extra={"plugin": plugin},
                            )

                    if callable(on_backup_start):
                        on_backup_start(name, temporary_path, exclude)

                    with zipfile.ZipFile(
                        temporary_path, mode="w", compression=compression, allowZip64=True
                    ) as zip:

                        def add_to_zip(source, target, ignored=None):
                            if ignored is None:
                                ignored = []

                            if source in ignored:
                                return

                            if os.path.isdir(source):
                                for entry in os.scandir(source):
                                    add_to_zip(
                                        entry.path,
                                        os.path.join(target, entry.name),
                                        ignored=ignored,
                                    )
                            elif os.path.isfile(source):
                                zip.write(source, arcname=target)

                        # add metadata
                        metadata = {
                            "version": get_octoprint_version_string(),
                            "excludes": exclude,
                        }
                        zip.writestr("metadata.json", json.dumps(metadata))

                        # backup current config file
                        add_to_zip(
                            configfile,
                            "basedir/config.yaml",
                            ignored=[
                                own_folder,
                            ],
                        )

                        # backup configured folder paths
                        for folder in default_settings["folder"].keys():
                            if folder in exclude or folder in exclude_by_default:
                                continue

                            add_to_zip(
                                settings.global_get_basefolder(folder),
                                "basedir/" + folder.replace("_", "/"),
                                ignored=[
                                    own_folder,
                                ]
                                + additional_excludes,
                            )

                        # backup anything else that might be lying around in our basedir
                        add_to_zip(
                            basedir,
                            "basedir",
                            ignored=defaults
                            + [
                                own_folder,
                            ]
                            + additional_excludes,
                        )

                        # add list of installed plugins
                        helpers = plugin_manager.get_helpers(
                            "pluginmanager", "generate_plugins_json"
                        )
                        if helpers and "generate_plugins_json" in helpers:
                            plugins = helpers["generate_plugins_json"](
                                settings=settings, plugin_manager=plugin_manager
                            )

                            if len(plugins):
                                zip.writestr("plugin_list.json", json.dumps(plugins))

                    shutil.move(temporary_path, final_path)

                    if callable(on_backup_done):
                        on_backup_done(name, final_path, exclude)

                    return final_path

                except Exception:
                    backup_error = True
                    raise
                finally:
                    for plugin, hook in plugin_manager.get_hooks(
                        "octoprint.plugin.backup.after_backup"
                    ).items():
                        try:
                            hook(error=backup_error)
                        except Exception:
                            logger.exception(
                                f"Error while running after_backup hook from plugin {plugin}",
                                extra={"plugin": plugin},
                            )

            except Exception as exc:  # noqa: F841
                # TODO py3: use the exception, not sys.exc_info()
                if callable(on_backup_error):
                    exc_info = sys.exc_info()
                    try:
                        on_backup_error(name, exc_info)
                    finally:
                        del exc_info
                raise

    @classmethod
    def _restore_backup(
        cls,
        path,
        settings=None,
        plugin_manager=None,
        datafolder=None,
        on_install_plugins=None,
        on_report_unknown_plugins=None,
        on_invalid_backup=None,
        on_log_progress=None,
        on_log_error=None,
        on_restore_start=None,
        on_restore_done=None,
        on_restore_failed=None,
    ):
        if not cls._restore_supported(settings):
            if callable(on_log_error):
                on_log_error("Restore is not supported on this operating system")
            if callable(on_restore_failed):
                on_restore_failed(path)
            return False

        logger = logging.getLogger(__name__)

        restart_command = settings.global_get(
            ["server", "commands", "serverRestartCommand"]
        )

        basedir = settings._basedir
        cls._clean_dir_backup(basedir, on_log_progress=on_log_progress)

        repo_url = settings.global_get(["plugins", "pluginmanager", "repository"])
        if not repo_url:
            repo_url = DEFAULT_PLUGIN_REPOSITORY

        plugin_repo = cls._get_plugin_repository_data(repo_url)

        with cls._restore_in_progress:
            restore_error = False
            try:
                for plugin, hook in plugin_manager.get_hooks(
                    "octoprint.plugin.backup.before_restore"
                ).items():
                    try:
                        hook()
                    except Exception:
                        logger.exception(
                            f"Error while running before_restore hook from plugin {plugin}",
                            extra={"plugin": plugin},
                        )

                if callable(on_restore_start):
                    on_restore_start(path)

                try:
                    with zipfile.ZipFile(path, "r") as zip:
                        # read metadata
                        try:
                            metadata_zipinfo = zip.getinfo("metadata.json")
                        except KeyError:
                            if callable(on_invalid_backup):
                                on_invalid_backup(
                                    "Not an OctoPrint backup, lacks metadata.json"
                                )
                            if callable(on_restore_failed):
                                on_restore_failed(path)
                            return False

                        metadata_bytes = zip.read(metadata_zipinfo)
                        metadata = json.loads(metadata_bytes)

                        backup_version = get_comparable_version(
                            metadata["version"], cut=1
                        )
                        if backup_version > get_octoprint_version(cut=1):
                            if callable(on_invalid_backup):
                                on_invalid_backup(
                                    "Backup is from a newer version of OctoPrint and cannot be applied"
                                )
                            if callable(on_restore_failed):
                                on_restore_failed(path)
                            return False

                        # unzip to temporary folder
                        temp = tempfile.mkdtemp()
                        try:
                            if callable(on_log_progress):
                                on_log_progress(f"Unpacking backup to {temp}...")

                            abstemp = os.path.abspath(temp)
                            dirs = {}
                            for member in zip.infolist():
                                abspath = os.path.abspath(
                                    os.path.join(temp, member.filename)
                                )
                                if abspath.startswith(abstemp):
                                    date_time = time.mktime(member.date_time + (0, 0, -1))

                                    zip.extract(member, temp)

                                    if os.path.isdir(abspath):
                                        dirs[abspath] = date_time
                                    else:
                                        os.utime(abspath, (date_time, date_time))

                            # set time on folders
                            for abspath, date_time in dirs.items():
                                os.utime(abspath, (date_time, date_time))

                            # sanity check
                            configfile = os.path.join(temp, "basedir", "config.yaml")
                            if not os.path.exists(configfile):
                                if callable(on_invalid_backup):
                                    on_invalid_backup("Backup lacks config.yaml")
                                if callable(on_restore_failed):
                                    on_restore_failed(path)
                                return False

                            configdata = yaml.load_from_file(path=configfile)

                            userfile = os.path.join(temp, "basedir", "users.yaml")
                            if not os.path.exists(userfile):
                                if callable(on_invalid_backup):
                                    on_invalid_backup("Backup lacks users.yaml")
                                if callable(on_restore_failed):
                                    on_restore_failed(path)
                                return False

                            if callable(on_log_progress):
                                on_log_progress("Unpacked")

                            # install available plugins
                            plugins = []
                            plugin_list_file = os.path.join(temp, "plugin_list.json")
                            if os.path.exists(plugin_list_file):
                                with open(
                                    os.path.join(temp, "plugin_list.json"), "rb"
                                ) as f:
                                    plugins = json.load(f)

                            known_plugins = []
                            unknown_plugins = []
                            if plugins:
                                if plugin_repo:
                                    for plugin in plugins:
                                        if plugin["key"] in plugin_manager.plugins:
                                            # already installed
                                            continue

                                        if plugin["key"] in plugin_repo:
                                            # not installed, can be installed from repository url
                                            known_plugins.append(
                                                plugin_repo[plugin["key"]]
                                            )
                                        else:
                                            # not installed, not installable
                                            unknown_plugins.append(plugin)

                                else:
                                    # no repo, all plugins are not installable
                                    unknown_plugins = plugins

                                if callable(on_log_progress):
                                    if known_plugins:
                                        on_log_progress(
                                            "Known and installable plugins: {}".format(
                                                ", ".join(
                                                    map(lambda x: x["id"], known_plugins)
                                                )
                                            )
                                        )
                                    if unknown_plugins:
                                        on_log_progress(
                                            "Unknown plugins: {}".format(
                                                ", ".join(
                                                    map(
                                                        lambda x: x["key"],
                                                        unknown_plugins,
                                                    )
                                                )
                                            )
                                        )

                                if callable(on_install_plugins):
                                    on_install_plugins(known_plugins)

                                if callable(on_report_unknown_plugins):
                                    on_report_unknown_plugins(unknown_plugins)

                            # move config data
                            basedir_backup = basedir + ".bck"
                            basedir_extracted = os.path.join(temp, "basedir")

                            if callable(on_log_progress):
                                on_log_progress(
                                    f"Renaming {basedir} to {basedir_backup}..."
                                )
                            shutil.move(basedir, basedir_backup)

                            try:
                                if callable(on_log_progress):
                                    on_log_progress(
                                        f"Moving {basedir_extracted} to {basedir}..."
                                    )
                                shutil.move(basedir_extracted, basedir)
                            except Exception:
                                if callable(on_log_error):
                                    on_log_error(
                                        "Error while restoring config data",
                                        exc_info=sys.exc_info(),
                                    )
                                    on_log_error("Rolling back old config data")

                                shutil.move(basedir_backup, basedir)

                                if callable(on_restore_failed):
                                    on_restore_failed(path)
                                return False

                            if unknown_plugins:
                                if callable(on_log_progress):
                                    on_log_progress(
                                        "Writing info file about unknown plugins"
                                    )

                                if not os.path.isdir(datafolder):
                                    os.makedirs(datafolder)

                                unknown_plugins_path = os.path.join(
                                    datafolder, UNKNOWN_PLUGINS_FILE
                                )
                                try:
                                    with open(unknown_plugins_path, mode="wb") as f:
                                        f.write(to_bytes(json.dumps(unknown_plugins)))
                                except Exception:
                                    if callable(on_log_error):
                                        on_log_error(
                                            "Could not persist list of unknown plugins to {}".format(
                                                unknown_plugins_path
                                            ),
                                            exc_info=sys.exc_info(),
                                        )

                        finally:
                            for plugin, hook in plugin_manager.get_hooks(
                                "octoprint.plugin.backup.before_restore"
                            ).items():
                                try:
                                    hook(error=restore_error)
                                except Exception:
                                    logger.exception(
                                        f"Error while running after_restore hook from plugin {plugin}",
                                        extra={"plugin": plugin},
                                    )

                            if callable(on_log_progress):
                                on_log_progress("Removing temporary unpacked folder")
                            shutil.rmtree(temp)

                except Exception:
                    restore_error = True
                    exc_info = sys.exc_info()
                    try:
                        if callable(on_log_error):
                            on_log_error("Error while running restore", exc_info=exc_info)
                        if callable(on_restore_failed):
                            on_restore_failed(path)
                    finally:
                        del exc_info
                    return False

                finally:
                    # remove zip
                    if callable(on_log_progress):
                        on_log_progress("Removing temporary zip")
                    os.remove(path)

            finally:
                for plugin, hook in plugin_manager.get_hooks(
                    "octoprint.plugin.backup.after_restore"
                ).items():
                    try:
                        hook(error=restore_error)
                    except Exception:
                        logger.exception(
                            f"Error while running after_restore hook from plugin {plugin}",
                            extra={"plugin": plugin},
                        )

            # restart server
            if not restart_command:
                restart_command = (
                    configdata.get("server", {})
                    .get("commands", {})
                    .get("serverRestartCommand")
                )

            if restart_command:
                import sarge

                if callable(on_log_progress):
                    on_log_progress("Restarting...")
                if callable(on_restore_done):
                    on_restore_done(path)

                try:
                    sarge.run(restart_command, close_fds=True, async_=True)
                except Exception:
                    if callable(on_log_error):
                        on_log_error(
                            f"Error while restarting via command {restart_command}",
                            exc_info=sys.exc_info(),
                        )
                        on_log_error("Please restart OctoPrint manually")
                    return False

            else:
                if callable(on_restore_done):
                    on_restore_done(path)
                if callable(on_log_error):
                    on_log_error(
                        "No restart command configured. Please restart OctoPrint manually."
                    )

        return True

    @classmethod
    def _build_backup_filename(cls, settings):
        backup_prefix = cls._get_backup_prefix(settings)
        return "{}-backup-{}.zip".format(
            backup_prefix, time.strftime(BACKUP_DATE_TIME_FMT)
        )

    @classmethod
    def _get_backup_prefix(cls, settings):
        if settings.global_get(["appearance", "name"]) == "":
            backup_prefix = "octoprint"
        else:
            backup_prefix = settings.global_get(["appearance", "name"])
        return sanitize(backup_prefix)

    @classmethod
    def _restore_supported(cls, settings):
        return (
            is_os_compatible(["!windows"])
            and not settings.get_boolean(["restore_unsupported"])
            and os.environ.get("OCTOPRINT_BACKUP_RESTORE_UNSUPPORTED", False)
            not in valid_boolean_trues
        )

    @classmethod
    def _valid_backup(cls, path):
        if not path.endswith(".zip") or not zipfile.is_zipfile(path):
            return False

        try:
            with zipfile.ZipFile(path) as z:
                return "metadata.json" in z.namelist()
        except Exception:
            return False

    def _send_client_message(self, message, payload=None):
        if payload is None:
            payload = {}
        payload["type"] = message
        self._plugin_manager.send_plugin_message(self._identifier, payload)


class InsufficientSpace(Exception):
    pass


def _register_custom_events(*args, **kwargs):
    return ["backup_created"]


__plugin_name__ = "Backup & Restore"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Backup & restore your OctoPrint settings and data"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin you will no longer be able to backup "
    "& restore your OctoPrint settings and data."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = BackupPlugin()
__plugin_hooks__ = {
    "octoprint.server.http.routes": __plugin_implementation__.route_hook,
    "octoprint.server.http.bodysize": __plugin_implementation__.bodysize_hook,
    "octoprint.cli.commands": __plugin_implementation__.cli_commands_hook,
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
    "octoprint.events.register_custom_events": _register_custom_events,
    "octoprint.server.sockjs.emit": __plugin_implementation__.socket_emit_hook,
}
__plugin_helpers__ = {
    "create_backup": __plugin_implementation__.create_backup_helper,
    "delete_backup": __plugin_implementation__.delete_backup_helper,
}



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2016 The OctoPrint Project - Released under terms of the AGPLv3 License"


import calendar
import logging
import os
import re
import threading
import time
from collections import OrderedDict

import feedparser
import flask
from flask_babel import gettext

import octoprint.plugin
from octoprint import __version__ as OCTOPRINT_VERSION
from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions
from octoprint.server.util.flask import (
    check_etag,
    no_firstrun_access,
    with_revalidation_checking,
)
from octoprint.util import count, deserialize, serialize, utmify
from octoprint.util.text import sanitize


class AnnouncementPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.EventHandlerPlugin,
):
    # noinspection PyMissingConstructor
    def __init__(self):
        self._cached_channel_configs = None
        self._cached_channel_configs_mutex = threading.RLock()

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "READ",
                "name": "Read announcements",
                "description": gettext("Allows to read announcements"),
                "default_groups": [ADMIN_GROUP],
                "roles": ["read"],
            },
            {
                "key": "MANAGE",
                "name": "Manage announcement subscriptions",
                "description": gettext(
                    'Allows to manage announcement subscriptions. Includes "Read announcements" '
                    "permission"
                ),
                "default_groups": [ADMIN_GROUP],
                "roles": ["manage"],
                "permissions": ["PLUGIN_ANNOUNCEMENTS_READ"],
            },
        ]

    # StartupPlugin

    def on_after_startup(self):
        # decouple channel fetching from server startup
        def fetch_data():
            self._fetch_all_channels()

        thread = threading.Thread(target=fetch_data)
        thread.daemon = True
        thread.start()

    # SettingsPlugin

    def get_settings_defaults(self):
        settings = {
            "channels": {
                "_important": {
                    "name": "Important Announcements",
                    "description": "Important announcements about OctoPrint.",
                    "priority": 1,
                    "type": "rss",
                    "url": "https://octoprint.org/feeds/important.xml",
                },
                "_releases": {
                    "name": "Release Announcements",
                    "description": "Announcements of new releases and release candidates of OctoPrint.",
                    "priority": 2,
                    "type": "rss",
                    "url": "https://octoprint.org/feeds/releases.xml",
                },
                "_blog": {
                    "name": "On the OctoBlog",
                    "description": "Development news, community spotlights, OctoPrint On Air episodes and more from the official OctoBlog.",
                    "priority": 2,
                    "type": "rss",
                    "url": "https://octoprint.org/feeds/octoblog.xml",
                },
                "_plugins": {
                    "name": "New Plugins in the Repository",
                    "description": "Announcements of new plugins released on the official Plugin Repository.",
                    "priority": 2,
                    "type": "rss",
                    "url": "https://plugins.octoprint.org/feed.xml",
                },
                "_octopi": {
                    "name": "OctoPi News",
                    "description": "News around OctoPi, the Raspberry Pi image including OctoPrint.",
                    "priority": 2,
                    "type": "rss",
                    "url": "https://octoprint.org/feeds/octopi.xml",
                },
            },
            "enabled_channels": [],
            "forced_channels": ["_important"],
            "channel_order": ["_important", "_releases", "_blog", "_plugins", "_octopi"],
            "ttl": 6 * 60,
            "display_limit": 3,
            "summary_limit": 300,
        }
        settings["enabled_channels"] = list(settings["channels"].keys())
        return settings

    def get_settings_version(self):
        return 1

    def on_settings_migrate(self, target, current):
        if current is None:
            # first version had different default feeds and only _important enabled by default
            channels = self._settings.get(["channels"])
            if "_news" in channels:
                del channels["_news"]
            if "_spotlight" in channels:
                del channels["_spotlight"]
            self._settings.set(["channels"], channels)

            enabled = self._settings.get(["enabled_channels"])
            add_blog = False
            if "_news" in enabled:
                add_blog = True
                enabled.remove("_news")
            if "_spotlight" in enabled:
                add_blog = True
                enabled.remove("_spotlight")
            if add_blog and "_blog" not in enabled:
                enabled.append("_blog")
            self._settings.set(["enabled_channels"], enabled)

    # AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/announcements.js"],
            "less": ["less/announcements.less"],
            "css": ["css/announcements.css"],
        }

    # Template Plugin

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "name": gettext("Announcements"),
                "template": "announcements_settings.jinja2",
                "custom_bindings": True,
            },
            {
                "type": "navbar",
                "template": "announcements_navbar.jinja2",
                "styles": ["display: none"],
                "data_bind": "visible: loginState.hasPermission(access.permissions.PLUGIN_ANNOUNCEMENTS_READ)",
            },
        ]

    # Blueprint Plugin

    @octoprint.plugin.BlueprintPlugin.route("/channels", methods=["GET"])
    @no_firstrun_access
    @Permissions.PLUGIN_ANNOUNCEMENTS_READ.require(403)
    def get_channel_data(self):
        from octoprint.settings import valid_boolean_trues

        result = []

        force = flask.request.values.get("force", "false") in valid_boolean_trues

        enabled = self._settings.get(["enabled_channels"])
        forced = self._settings.get(["forced_channels"])

        channel_configs = self._get_channel_configs(force=force)

        def view():
            channel_data = self._fetch_all_channels(force=force)

            for key, data in channel_configs.items():
                read_until = channel_configs[key].get("read_until", None)
                entries = sorted(
                    self._to_internal_feed(
                        channel_data.get(key, []), read_until=read_until
                    ),
                    key=lambda e: e["published"],
                    reverse=True,
                )
                unread = count(filter(lambda e: not e["read"], entries))

                if read_until is None and entries:
                    last = entries[0]["published"]
                    self._mark_read_until(key, last)

                result.append(
                    {
                        "key": key,
                        "channel": data["name"],
                        "url": data["url"],
                        "description": data.get("description", ""),
                        "priority": data.get("priority", 2),
                        "enabled": key in enabled or key in forced,
                        "forced": key in forced,
                        "data": entries,
                        "unread": unread,
                    }
                )

            return flask.jsonify(channels=result)

        def etag():
            import hashlib

            hash = hashlib.sha1()

            def hash_update(value):
                hash.update(value.encode("utf-8"))

            hash_update(repr(sorted(enabled)))
            hash_update(repr(sorted(forced)))
            hash_update(OCTOPRINT_VERSION)

            for channel in sorted(channel_configs.keys()):
                hash_update(repr(channel_configs[channel]))
                channel_data = self._get_channel_data_from_cache(
                    channel, channel_configs[channel]
                )
                hash_update(repr(channel_data))

            return hash.hexdigest()

        # noinspection PyShadowingNames
        def condition(lm, etag):
            return check_etag(etag)

        return with_revalidation_checking(
            etag_factory=lambda *args, **kwargs: etag(),
            condition=lambda lm, etag: condition(lm, etag),
            unless=lambda: force,
        )(view)()

    @octoprint.plugin.BlueprintPlugin.route("/channels/<channel>", methods=["POST"])
    @no_firstrun_access
    @Permissions.PLUGIN_ANNOUNCEMENTS_READ.require(403)
    def channel_command(self, channel):
        from octoprint.server import NO_CONTENT
        from octoprint.server.util.flask import get_json_command_from_request

        valid_commands = {"read": ["until"], "toggle": []}

        command, data, response = get_json_command_from_request(
            flask.request, valid_commands=valid_commands
        )
        if response is not None:
            return response

        if command == "read":
            until = data["until"]
            self._mark_read_until(channel, until)

        elif command == "toggle":
            if not Permissions.PLUGIN_ANNOUNCEMENTS_MANAGE.can():
                flask.abort(403)
            self._toggle(channel)

        return NO_CONTENT

    @octoprint.plugin.BlueprintPlugin.route("/channels", methods=["POST"])
    @no_firstrun_access
    @Permissions.PLUGIN_ANNOUNCEMENTS_READ.require(403)
    def channels_command(self):
        from octoprint.server import NO_CONTENT
        from octoprint.server.util.flask import get_json_command_from_request

        valid_commands = {"read": []}

        command, _, response = get_json_command_from_request(
            flask.request, valid_commands=valid_commands
        )
        if response is not None:
            return response

        if command == "read":
            channel_data = self._fetch_all_channels()
            for key, entries in channel_data.items():
                if not entries:
                    continue
                last = sorted(
                    self._to_internal_feed(entries),
                    key=lambda x: x["published"],
                    reverse=True,
                )[0]["published"]
                self._mark_read_until(key, last)

        return NO_CONTENT

    def is_blueprint_protected(self):
        return False

    def is_blueprint_csrf_protected(self):
        return True

    ##~~ EventHandlerPlugin

    def on_event(self, event, payload):
        from octoprint.events import Events

        if (
            event != Events.CONNECTIVITY_CHANGED
            or not payload
            or not payload.get("new", False)
        ):
            return
        self._fetch_all_channels_async()

    # Internal Tools

    def _mark_read_until(self, channel, until):
        """Set read_until timestamp of a channel."""

        current_read_until = None
        channel_data = self._settings.get(["channels", channel], merged=True)
        if channel_data:
            current_read_until = channel_data.get("read_until", None)

        defaults = {"plugins": {"announcements": {"channels": {}}}}
        defaults["plugins"]["announcements"]["channels"][channel] = {
            "read_until": current_read_until
        }

        with self._cached_channel_configs_mutex:
            self._settings.set(
                ["channels", channel, "read_until"], until, defaults=defaults
            )
            self._settings.save()
            self._cached_channel_configs = None

    def _toggle(self, channel):
        """Toggle enable/disabled state of a channel."""

        enabled_channels = list(self._settings.get(["enabled_channels"]))

        if channel in enabled_channels:
            enabled_channels.remove(channel)
        else:
            enabled_channels.append(channel)

        self._settings.set(["enabled_channels"], enabled_channels)
        self._settings.save()

    def _get_channel_configs(self, force=False):
        """Retrieve all channel configs with sanitized keys."""

        with self._cached_channel_configs_mutex:
            if self._cached_channel_configs is None or force:
                configs = self._settings.get(["channels"], merged=True)
                order = self._settings.get(["channel_order"])
                all_keys = order + [
                    key for key in sorted(configs.keys()) if key not in order
                ]

                result = OrderedDict()
                for key in all_keys:
                    config = configs.get(key)
                    if config is None or "url" not in config or "name" not in config:
                        # strip invalid entries
                        continue
                    result[sanitize(key)] = config
                self._cached_channel_configs = result
        return self._cached_channel_configs

    def _get_channel_config(self, key, force=False):
        """Retrieve specific channel config for channel."""

        safe_key = sanitize(key)
        return self._get_channel_configs(force=force).get(safe_key)

    def _fetch_all_channels_async(self, force=False):
        thread = threading.Thread(
            target=self._fetch_all_channels, kwargs={"force": force}
        )
        thread.daemon = True
        thread.start()

    def _fetch_all_channels(self, force=False):
        """Fetch all channel feeds from cache or network."""

        channels = self._get_channel_configs(force=force)
        enabled = self._settings.get(["enabled_channels"])
        forced = self._settings.get(["forced_channels"])

        all_channels = {}
        for key, config in channels.items():
            if key not in enabled and key not in forced:
                continue

            if "url" not in config:
                continue

            data = self._get_channel_data(key, config, force=force)
            if data is not None:
                all_channels[key] = data

        return all_channels

    def _get_channel_data(self, key, config, force=False):
        """Fetch individual channel feed from cache/network."""

        data = None

        if not force:
            # we may use the cache, see if we have something in there
            data = self._get_channel_data_from_cache(key, config)

        if data is None:
            # cache not allowed or empty, fetch from network
            if self._connectivity_checker.check_immediately():
                data = self._get_channel_data_from_network(key, config)
            else:
                self._logger.info(
                    "Looks like we are offline, can't fetch announcements for channel {} from network".format(
                        key
                    )
                )

        return data

    def _get_channel_data_from_cache(self, key, config):
        """Fetch channel feed from cache."""

        channel_path = self._get_channel_cache_path(key)

        if os.path.exists(channel_path):
            if "ttl" in config and isinstance(config["ttl"], int):
                ttl = config["ttl"]
            else:
                ttl = self._settings.get_int(["ttl"])

            ttl *= 60
            now = time.time()
            if now > os.stat(channel_path).st_mtime + ttl:
                return None

            try:
                d = deserialize(channel_path)
                self._logger.debug(f"Loaded channel {key} from cache at {channel_path}")
                return d
            except Exception:
                if self._logger.isEnabledFor(logging.DEBUG):
                    self._logger.exception(
                        f"Could not read channel {key} from cache at {channel_path}"
                    )

                try:
                    os.remove(channel_path)
                except OSError:
                    pass
                return None

    def _get_channel_data_from_network(self, key, config):
        """Fetch channel feed from network."""

        import requests

        url = config["url"]
        try:
            start = time.monotonic()
            r = requests.get(url, timeout=3.05)
            r.raise_for_status()
            self._logger.info(
                "Loaded channel {} from {} in {:.2}s".format(
                    key, config["url"], time.monotonic() - start
                )
            )
        except Exception:
            self._logger.exception(
                "Could not fetch channel {} from {}".format(key, config["url"])
            )
            return None

        response = feedparser.parse(r.text)
        try:
            serialize(self._get_channel_cache_path(key), response)
        except Exception:
            self._logger.exception(f"Failed to cache data for channel {key}")
        return response

    def _to_internal_feed(self, feed, read_until=None):
        """Convert feed to internal data structure."""

        result = []
        if "entries" in feed:
            for entry in feed["entries"]:
                try:
                    internal_entry = self._to_internal_entry(entry, read_until=read_until)
                    if internal_entry:
                        result.append(internal_entry)
                except Exception:
                    self._logger.exception(
                        "Error while converting entry from feed, skipping it"
                    )
        return result

    def _to_internal_entry(self, entry, read_until=None):
        """Convert feed entries to internal data structure."""

        timestamp = entry.get("published_parsed", None)
        if timestamp is None:
            timestamp = entry.get("updated_parsed", None)
        if timestamp is None:
            return None

        published = calendar.timegm(timestamp)

        read = True
        if read_until is not None:
            read = published <= read_until

        return {
            "title": entry["title"],
            "title_without_tags": _strip_tags(entry["title"]),
            "summary": _lazy_images(entry["summary"]),
            "summary_without_images": _strip_images(entry["summary"]),
            "published": published,
            "link": utmify(
                entry["link"],
                source="octoprint",
                medium="announcements",
                content=OCTOPRINT_VERSION,
            ),
            "read": read,
        }

    def _get_channel_cache_path(self, key):
        """Retrieve cache path for channel key."""

        safe_key = sanitize(key)
        return os.path.join(self.get_plugin_data_folder(), f"{safe_key}.cache")


_image_tag_re = re.compile(r"<img.*?/?>")


def _strip_images(text):
    """
    >>> _strip_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>")
    "<a href='test.html'>I'm a link</a> and this is an image: "
    >>> _strip_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">")
    'One  and two  and three  and four '
    >>> _strip_images("No images here")
    'No images here'
    """
    return _image_tag_re.sub("", text)


def _replace_images(text, callback):
    """
    >>> callback = lambda img: "foobar"
    >>> _replace_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>", callback)
    "<a href='test.html'>I'm a link</a> and this is an image: foobar"
    >>> _replace_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">", callback)
    'One foobar and two foobar and three foobar and four foobar'
    """
    result = text
    for match in _image_tag_re.finditer(text):
        tag = match.group(0)
        replaced = callback(tag)
        result = result.replace(tag, replaced)
    return result


_image_src_re = re.compile(r'src=(?P<quote>[\'"]*)(?P<src>.*?)(?P=quote)(?=\s+|>)')


def _lazy_images(text, placeholder=None):
    """
    >>> _lazy_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>")
    '<a href=\\'test.html\\'>I\\'m a link</a> and this is an image: <img src="data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7" data-src=\\'foo.jpg\\' alt=\\'foo\\'>'
    >>> _lazy_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>", placeholder="ph.png")
    '<a href=\\'test.html\\'>I\\'m a link</a> and this is an image: <img src="ph.png" data-src=\\'foo.jpg\\' alt=\\'foo\\'>'
    >>> _lazy_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">", placeholder="ph.png")
    'One <img src="ph.png" data-src="one.jpg"> and two <img src="ph.png" data-src=\\'two.jpg\\' > and three <img src="ph.png" data-src=three.jpg> and four <img src="ph.png" data-src="four.png" alt="four">'
    >>> _lazy_images("No images here")
    'No images here'
    """
    if placeholder is None:
        # 1px transparent gif
        placeholder = "data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7"

    def callback(img_tag):
        match = _image_src_re.search(img_tag)
        if match is not None:
            src = match.group("src")
            quote = match.group("quote")
            quoted_src = quote + src + quote
            img_tag = img_tag.replace(
                match.group(0), f'src="{placeholder}" data-src={quoted_src}'
            )
        return img_tag

    return _replace_images(text, callback)


def _strip_tags(text):
    """
    >>> _strip_tags("<a href='test.html'>Hello world</a>&lt;img src='foo.jpg'&gt;")
    "Hello world&lt;img src='foo.jpg'&gt;"
    >>> _strip_tags("&#62; &#x3E; Foo")
    '&#62; &#x3E; Foo'
    """
    from html.parser import HTMLParser

    class TagStripper(HTMLParser):
        def __init__(self, **kw):
            HTMLParser.__init__(self, **kw)
            self._fed = []

        def handle_data(self, data):
            self._fed.append(data)

        def handle_entityref(self, ref):
            self._fed.append(f"&{ref};")

        def handle_charref(self, ref):
            self._fed.append(f"&#{ref};")

        def get_data(self):
            return "".join(self._fed)

    tag_stripper = TagStripper(convert_charrefs=False)
    tag_stripper.feed(text)
    return tag_stripper.get_data()


__plugin_name__ = "Announcement Plugin"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Announcements all around OctoPrint"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin you might miss important announcements "
    "regarding security or other critical issues concerning OctoPrint."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = AnnouncementPlugin()

__plugin_hooks__ = {
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
}



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import threading
import time

from octoprint.util.net import resolve_host, server_reachable


class ConnectivityChecker:
    """
    Regularly checks for online connectivity.

    Tries to open a connection to the provided ``host`` and ``port`` every ``interval``
    seconds and sets the ``online`` status accordingly.

    If a ``name``is provided, also tries to resolve that name to a valid IP address
    during connectivity check and only set ``online`` to ``True`` if that succeeds as well.
    """

    def __init__(self, interval, host, port, name=None, enabled=True, on_change=None):
        self._interval = interval
        self._host = host
        self._port = port
        self._name = name
        self._enabled = enabled
        self._on_change = on_change

        self._logger = logging.getLogger(__name__ + ".connectivity_checker")

        # we initialize the online flag to True if we are not enabled (we don't know any better
        # but these days it's probably a sane default)
        self._connection_working = not self._enabled
        self._resolution_working = not self._enabled or self._name is None

        self._check_worker = None
        self._check_mutex = threading.RLock()

        self._run()

    @property
    def online(self):
        """Current online status, True if online, False if offline."""
        with self._check_mutex:
            return self._online

    @property
    def _online(self):
        return self._connection_working and self._resolution_working

    @property
    def host(self):
        """DNS host to query."""
        with self._check_mutex:
            return self._host

    @host.setter
    def host(self, value):
        with self._check_mutex:
            self._host = value

    @property
    def port(self):
        """DNS port to query."""
        with self._check_mutex:
            return self._port

    @port.setter
    def port(self, value):
        with self._check_mutex:
            self._port = value

    @property
    def name(self):
        with self._check_mutex:
            return self._name

    @name.setter
    def name(self, value):
        with self._check_mutex:
            self._name = value

    @property
    def interval(self):
        """Interval between consecutive automatic checks."""
        return self._interval

    @interval.setter
    def interval(self, value):
        self._interval = value

    @property
    def enabled(self):
        """Whether the check is enabled or not."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        with self._check_mutex:
            old_enabled = self._enabled
            self._enabled = value

            if not self._enabled:
                if self._check_worker is not None:
                    self._check_worker.cancel()

                old_value = self._online
                self._connection_working = self._resolution_working = True

                if old_value != self._online:
                    self._trigger_change(old_value, self._online)

            elif self._enabled and not old_enabled:
                self._run()

    def check_immediately(self):
        """Check immediately and return result."""
        with self._check_mutex:
            self._perform_check()
            return self.online

    def log_full_report(self):
        with self._check_mutex:
            self._logger.info(
                "Connectivity state is currently: {}".format(
                    self._map_online(self.online)
                )
            )
            self.log_details()

    def log_details(self):
        self._logger.info(
            "Connecting to {}:{} is {}".format(
                self._host,
                self._port,
                self._map_working(self._connection_working),
            )
        )
        if self._name:
            self._logger.info(
                "Resolving {} is {}".format(
                    self._name,
                    self._map_working(self._resolution_working),
                )
            )

    def as_dict(self):
        result = {
            "online": self.online,
            "enabled": self.enabled,
            "connection_ok": self._connection_working,
            "connection_check": f"{self._host}:{self._port}",
        }
        if self._name:
            result.update(
                resolution_ok=self._resolution_working, resolution_check=self._name
            )
        return result

    def _run(self):
        from octoprint.util import RepeatedTimer

        if not self._enabled:
            return

        if self._check_worker is not None:
            self._check_worker.cancel()

        self._check_worker = RepeatedTimer(
            self._interval, self._perform_check, run_first=True
        )
        self._check_worker.start()

    def _perform_check(self):
        if not self._enabled:
            return

        with self._check_mutex:
            self._logger.debug(
                "Checking against {}:{} if we are online...".format(
                    self._host, self._port
                )
            )

            old_value = self._online

            for _ in range(3):
                connection_working = server_reachable(self._host, port=self._port)

                if self._name:
                    if connection_working:
                        self._logger.debug(f"Checking if we can resolve {self._name}...")
                        resolution_working = len(resolve_host(self._name)) > 0
                    else:
                        resolution_working = False
                else:
                    resolution_working = True

                if not (connection_working and resolution_working):
                    # retry up to 3 times
                    time.sleep(1.0)
                    continue

                # all good
                break

            self._connection_working = connection_working
            self._resolution_working = resolution_working
            new_value = self._online

            if old_value != new_value:
                self._trigger_change(old_value, new_value)

    def _trigger_change(self, old_value, new_value):
        self._log_change_report(old_value, new_value, include_details=not new_value)
        if callable(self._on_change):
            self._on_change(
                old_value,
                new_value,
                connection_working=self._connection_working,
                resolution_working=self._resolution_working,
            )

    def _log_change_report(self, old_value, new_value, include_details=False):
        self._logger.info(
            "Connectivity changed from {} to {}".format(
                self._map_online(old_value),
                self._map_online(new_value),
            )
        )
        if include_details:
            self.log_details()

    def _map_working(self, value):
        return "working" if value else "not working"

    def _map_online(self, value):
        return "online" if value else "offline"



import logging

ADMIN_GROUP = "admins"
USER_GROUP = "users"
GUEST_GROUP = "guests"
READONLY_GROUP = "readonly"


def auth_log(message):
    logging.getLogger("AUTH").info(message)


login_mechanisms = login_mechanism_lut = {
    "http": "credentials",
    "autologin": "autologin",
    "remember_me": "Remember Me cookie",
    "basic_auth": "Basic Authorization header",
    "remote_user": "Remote User header",
}



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"

import datetime

LOCAL_TZ = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
UTC_TZ = datetime.timezone.utc


def is_timezone_aware(dt: datetime.datetime) -> bool:
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import List, Optional, Union

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class LayoutEnum(str, Enum):
    horizontal = "horizontal"
    vertical = "vertical"


class ControlSliderInputConfig(BaseModel):
    min: int = 0
    """Minimum value of the slider."""

    max: int = 255
    """Maximum value of the slider."""

    step: int = 1
    """Step size per slider tick."""


class ControlInputConfig(BaseModel):
    name: str
    """Name to display for the input field."""

    parameter: str
    """Internal parameter name for the input field, used as a placeholder in `command`/`commands`."""

    default: Union[str, int, float, bool]
    """Default value for the input field."""

    slider: Optional[ControlSliderInputConfig] = None
    """If this attribute is included, instead of an input field a slider control will be rendered."""


@with_attrs_docs
class ContainerConfig(BaseModel):
    children: "List[Union[ContainerConfig, ControlConfig]]" = []
    """A list of children controls or containers contained within this container."""

    name: Optional[str] = None
    """A name to display above the container, basically a section header."""

    layout: LayoutEnum = LayoutEnum.vertical
    """The layout to use for laying out the contained children, either from top to bottom (`vertical`) or from left to right (`horizontal`)."""


@with_attrs_docs
class ControlConfig(BaseModel):
    name: str
    """The name of the control, will be displayed either on the button if it's a control sending a command or as a label for controls which only display output."""

    command: Optional[str] = None
    """A single GCODE command to send to the printer. Will be rendered as a button which sends the command to the printer upon click. The button text will be the value of the `name` attribute. Mutually exclusive with `commands` and `script`. The rendered button be disabled if the printer is currently offline or printing or alternatively if the requirements defined via the `enabled` attribute are not met."""

    commands: Optional[List[str]] = None
    """A list of GCODE commands to send to the printer. Will be rendered as a button which sends the commands to the printer upon click. The button text will be the value of the `name` attribute. Mutually exclusive with `command` and `script`. The rendered button will be disabled if the printer is currently offline or printing or alternatively if the requirements defined via the `enabled` attribute are not met."""

    script: Optional[str] = None
    """The name of a full blown [GCODE script]() to send to the printer. Will be rendered as a button which sends the script to the printer upon click. The button text will be the value of the `name` attribute. Mutually exclusive with `command` and `commands`. The rendered button will be disabled if the printer is currently offline or printing or alternatively if the requirements defined via the `enabled` attribute are not met. Values of input parameters will be available in the template context under the `parameter` variable (e.g. an input parameter `speed` will be available in the script template as `parameter.speed`). On top of that all other variables defined in the [GCODE template context]() will be available."""

    javascript: Optional[str] = None
    """A JavaScript snippet to be executed when the button rendered for `command` or `commands` is clicked. This allows to override the direct sending of the command or commands to the printer with more sophisticated behaviour. The JavaScript snippet is `eval`'d and processed in a context where the control it is part of is provided as local variable `data` and the `ControlViewModel` is available as `self`."""

    additionalClasses: Optional[str] = None
    """Additional classes to apply to the button of a `command`, `commands`, `script` or `javascript` control, other than the default `btn`. Can be used to visually style the button, e.g. set to `btn-danger` to turn the button red."""

    enabled: Optional[str] = None
    """A JavaScript snippet returning either `true` or `false` determining whether the control should be enabled or not. This allows to override the default logic for the enable state of the control (disabled if printer is offline). The JavaScript snippet is `eval`'d and processed in a context where the control it is part of is provided as local variable `data` and the `ControlViewModel` is available as `self`."""

    input: Optional[List[ControlInputConfig]] = []
    """A list of definitions of input parameters for a `command` or `commands`, to be rendered as additional input fields. `command`/`commands` may contain placeholders to be replaced by the values obtained from the user for the defined input fields."""

    regex: Optional[str] = None
    """A [regular expression <re-syntax>](https://docs.python.org/3/library/re.html#regular-expression-syntax) to match against lines received from the printer to retrieve information from it (e.g. specific output). Together with `template` this allows rendition of received data from the printer within the UI."""

    template: Optional[str] = None
    r"""A template to use for rendering the match of `regex`. May contain placeholders in [Python Format String Syntax](https://docs.python.org/3/library/string.html#formatstrings) for either named groups within the regex (e.g. `Temperature: {temperature}` for a regex `T:\s*(?P<temperature>\d+(\.\d*)`) or positional groups within the regex (e.g. `Position: X={0}, Y={1}, Z={2}, E={3}` for a regex `X:([0-9.]+) Y:([0-9.]+) Z:([0-9.]+) E:([0-9.]+)`)."""

    confirm: Optional[str] = None
    """A text to display to the user to confirm his button press. Can be used with sensitive custom controls like changing EEPROM values in order to prevent accidental clicks. The text will be displayed in a confirmation dialog."""
